@@ -60,6 +60,7 @@ public:
 	void   printHelp()    const;
 	void   printVersion() const;
 	void   printTemplate()const;
+	void   printWarnings()const;
 private:
 	Application();
 	Application(const Application&);
@@ -70,6 +71,14 @@ private:
 		claspre_.initOptions(root);
 		cmdOpts_.setConfig(&config_);
 		cmdOpts_.initOptions(root, hidden);
+		// DEPRECATED options - remove in next version
+		ProgramOptions::OptionGroup deprecated("Clasp - DEPRECATED options");
+		deprecated.addOptions()		
+			("brave"              , ProgramOptions::flag(brave_)    , "Use --enum-mode=brave")
+			("cautious"           , ProgramOptions::flag(cautious_) , "Use --enum-mode=cautious")
+			("solution-recording" , ProgramOptions::flag(solu_)     , "Use --enum-mode=record\n")
+		;
+		root.add(deprecated);
 		ProgramOptions::OptionGroup basic("Basic Options");
 		basic.addOptions()
 			("lemma-out" , ProgramOptions::storeTo(lemmaOut_), "Write learnt lemmas to %A on exit", "<file>")
@@ -81,7 +90,8 @@ private:
 		o.value()->defaultValue("1");
 	}
 	bool validateOptions(const ProgramOptions::ParsedOptions& vm, Messages& m) {
-		return claspre_.validateOptions(vm) && cmdOpts_.validateOptions(vm, m);
+		return claspre_.validateOptions(vm) && cmdOpts_.validateOptions(vm, m)
+			&& validateDeprecated(vm);
 	}
 	// -------------------------------------------------------------------------------------------
 	// ClaspFacade::Callback interface
@@ -96,6 +106,7 @@ private:
 	void readLemmas();
 	void writeLemmas();
 	int  exception(int status, const char* what);
+	bool validateDeprecated(const ProgramOptions::ParsedOptions& vm);
 	// -------------------------------------------------------------------------------------------  
 	// Status information & output
 	void configureOutput(Input::Format f);
@@ -114,6 +125,10 @@ private:
 	ClaspFacade*                  facade_;
 	std::auto_ptr<OutputFormat>   out_;
 	volatile sig_atomic_t         stopped_;
+	// DEPRECATED options - remove in next version
+	bool                          brave_;
+	bool                          cautious_;
+	bool                          solu_;
 };
 
 }
