@@ -88,21 +88,20 @@ void GeneralOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpt
 	OptionGroup threadOps("Clasp - Thread Options");
 	ThreadOptions* thread = &config->thread;
 	threadOps.addOptions() 
-		("create-template,g" , flag(thread->genTemplate), "Generate a template portfolio file and exit")
 		("threads,t", lazy<int>(this, &GeneralOptions::mapThreads)->defaultValue("1")->state(Value::value_defaulted), 
 			"Set number of threads to use\n"
 			"      Default: %D / Valid: 0 <= %A < 64", "<n>")
 		("portfolio,p", storeTo(thread->portfolio), 
-			"Portfolio file to use for threads", "<file>")
-		("force-gp", flag(thread->forceGP), 
-			"Force guiding path scheme even if portfolio is used")
-		("copy-problem", flag(thread->distribute.copyProblem), 
-			"Copy (instead of share) problem between threads\n")
+		 "Use portfolio to configure threads\n"
+		 "      Valid: %A or 'default'", "<file>")
+		("create-template,g" , flag(thread->genTemplate), "Print a template portfolio and exit\n")
+		("force-gp", flag(thread->forceGP), "Force guiding path scheme even if portfolio is used")
 		("global-restarts", custom(this, &GeneralOptions::mapCustom)->defaultValue("0")->state(Value::value_defaulted),
 		 "Configure global restart policy\n"
 		 "      Default: %D / Valid: <n>[,strat>]\n"
 		 "        <n> : Maximal number of global restarts (0=disable)\n"
 		 "     <strat>: Restart strategy (see --restarts)\n", "<arg>")
+		("copy-problem", flag(thread->distribute.copyProblem), "Copy (instead of share) problem between threads")
 		("distribute", custom(this, &GeneralOptions::mapCustom)->defaultValue("no")->state(Value::value_defaulted),
 			"Configure nogood distribution\n"
 			"      Default: %D\n"
@@ -123,24 +122,29 @@ void GeneralOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpt
 		("number,n", storeTo(global->enumerate.numModels)->defaultValue("1")->state(Value::value_defaulted), 
 			"Enumerate at most %A models (0 for all)\n"
 			"      Default: %D (0 when optimizing/computing consequences)", "<n>")
-		("solution-recording", flag(global->enumerate.record)        , "Add conflicts for computed models")
-		("restart-on-model"  , flag(global->enumerate.restartOnModel), "Restart (instead of backtrack) after each model")
+		("enum-mode", storeTo(global->enumerate.mode, &GeneralOptions::parseEnumMode)->defaultValue("auto")->state(Value::value_defaulted), 
+			"Configure enumeration algorithm\n"
+			"      Default: %D\n"
+			"      Valid:   auto, backtrack, record, brave, cautious\n"
+			"        auto    : Automatically select between bt and record\n"
+			"        bt      : Backtrack decision literals from solutions\n"
+			"        record  : Add nogoods for computed solutions\n"
+			"        brave   : Compute brave consequences\n"
+			"        cautious: Compute cautious consequences")
+		("restart-on-model"  , flag(global->enumerate.restartOnModel), "Restart after each model")
 		("project"           , flag(global->enumerate.project)       , "Project models to named atoms in enumeration mode")
-		("brave"             , flag(global->enumerate.brave)         , "Compute brave consequences")
-		("cautious"          , flag(global->enumerate.cautious)      , "Compute cautious consequences")
+		("supp-models",flag(global->eq.noSCC), "Compute supported models (no unfounded set check)\n")
+		
 		("pre" , flag(global->enumerate.onlyPre), "Run ASP preprocessor and exit\n")
  
 		("opt-sat"    , flag(global->opt.sat), "Compute MaxSAT (only applicable for DIMACS input)")
 		("opt-ignore" , flag(global->opt.no) , "Ignore minimize statements")
-		("opt-heuristic" , storeTo(global->opt.heu)->implicitValue("1"),    "Use minimize in {1=sign, 2=model, 3=both} heuristics")
-		("opt-hierarch", storeTo(global->opt.hierarch)->implicitValue("1"), "Use {1=hierarchical, 2=step, 3=neg step} optimization")
-		("opt-all"    , lazy<vector<int> >(this, &GeneralOptions::mapVec), "Compute all models <= %A", "<opt>")
+		("opt-hierarch", storeTo(global->opt.hierarch)->implicitValue("1"), "Hierarchical opt. and {1=fixed, 2=inc, 3=dec} step")
+		("opt-all"    , lazy<vector<int> >(this, &GeneralOptions::mapVec), "Compute only models <= %A", "<opt>")
 		("opt-value"  , lazy<vector<int> >(this, &GeneralOptions::mapVec), 
 			"Initialize objective function(s)\n"
 			"      Valid:   <n1[,n2,n3,...]>\n")
 		
-		("supp-models",flag(global->eq.noSCC), "Compute supported models (no unfounded set check)\n")
-		
 		("trans-ext", storeTo(global->eq.erMode, &GeneralOptions::parseExtRule)->defaultValue("no")->state(Value::value_defaulted),
 			"Configure handling of Lparse-like extended rules\n"
 			"      Default: %D\n"
@@ -182,6 +186,17 @@ void GeneralOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpt
 	;
 }
 
+bool GeneralOptions::parseEnumMode(const std::string& s, GlobalOptions::EnumMode& m) {
+	std::string temp = toLower(s);
+	if      (temp == "auto")    { m = GlobalOptions::enum_auto; }
+	else if (temp == "bt")      { m = GlobalOptions::enum_bt; }
+	else if (temp == "record")  { m = GlobalOptions::enum_record; }
+	else if (temp == "brave")   { m = GlobalOptions::enum_brave; }
+	else if (temp == "cautious"){ m = GlobalOptions::enum_cautious; }
+	else                        { return false; }
+	return true;
+}
+
 bool GeneralOptions::parseExtRule(const std::string& s, ProgramBuilder::ExtendedRuleMode& i){
 	std::string temp = toLower(s);
 	bool b = temp == "all";
@@ -271,22 +286,9 @@ bool GeneralOptions::validateOptions(const ProgramOptions::ParsedOptions& vm, Me
 	if (vm.count("opt-value") && vm.count("opt-all")) {
 		m.warning.push_back("'opt-all' and 'opt-value' are mutually exclusive!");
 	}
-	if (config->opt.heu > 1 && !config->enumerate.restartOnModel) {
-		m.warning.push_back("'opt-heuristic' ignored because value requires '--restart-on-model'");
-		config->opt.heu -= 2;
-	}
 	if (config->eq.noSCC && vm.count("eq") == 0) {
 		config->eq.noEq();
 	}
-	if (!config->thread.portfolio.empty()) {
-		std::ifstream test(config->thread.portfolio.c_str());
-		if (!test) {
-			m.error = "Could not open portfolio file '";
-			m.error +=config->thread.portfolio;
-			m.error += "'";
-			return false;
-		}
-	}
 	return true;
 }
 
@@ -316,6 +318,7 @@ void SearchOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpti
 			"        Vsids  : Apply Chaff-like heuristic\n"
 			"        Unit   : Apply Smodels-like heuristic\n"
 			"        None   : Select the first free variable")
+		("opt-heuristic" , storeTo(local->solver().strategies().heuOpts)->implicitValue("1"),    "Use minimize in {1=sign, 2=model, 3=both} heuristics")
 		("rand-freq", lazy<double>(this, &SearchOptions::mapRandFreq)->defaultValue("0.0")->state(Value::value_defaulted), 
 			"Make random decisions with probability %A\n"
 			"      Default: %D\n"
@@ -339,14 +342,15 @@ void SearchOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpti
 	
 	OptionGroup lookback("Clasp - Lookback Options");
 	lookback.addOptions()
-		("restarts,r", custom(this, &SearchOptions::mapCustom)->defaultValue("100,1.5")->state(Value::value_defaulted),
+		("restarts,r", storeTo(local->solve.restart.sched, &SearchOptions::parseSchedule)->defaultValue("100,1.5")->state(Value::value_defaulted),
 			"Configure restart policy\n"
 			"      Default: %D\n"
-			"      Valid:   <n1[,n2,n3]> (<n1> >= 0, <n2>,<n3> > 0), no\n"
-			"        <n1>          : Run Luby et al.'s sequence with unit length <n1>\n"
-			"        <n1>,<n2>     : Run geometric sequence of <n1>*(<n2>^i) conflicts\n"
-			"        <n1>,<n2>,<n3>: Run Biere's inner-outer geometric sequence (<n3>=outer)\n"
-			"        <n1> = 0, no  : Disable restarts")
+			"      Valid:   [type,]<n1[,n2,n3]> (<n1> >= 0, <n2>,<n3> > 0), no\n"
+			"      [L,]<n1>          : Run Luby et al.'s sequence with unit length <n1>\n"
+			"      [*,]<n1>,<n2>     : Run geometric sequence of <n1>*(<n2>^i) conflicts\n"
+			"      +,<n1>,<n2>       : Run arithmetic sequence of <n1>+(<n2>*i) conflicts\n"
+			"      [+,]<n1>,<n2>,<n3>: Run Biere's inner-outer sequence (<n3>=outer)\n"
+			"          <n1> = 0, no  : Disable restarts")
 		("local-restarts"  , flag(local->solve.restart.local), "Enable Ryvchin et al.'s local restarts")
 		("bounded-restarts", flag(local->solve.restart.bounded), "Enable (bounded) restarts during model enumeration")
 		("reset-restarts",   flag(local->solve.restart.resetOnModel), "Reset restart strategy during model enumeration")
@@ -367,7 +371,7 @@ void SearchOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpti
 			"        <n1,n2,n3>: Store at most min(P/<n1>*(<n2>^i),P*<n3>) learnt nogoods,\n"
 			"                    P and i being initial problem size and number of restarts\n"
 			"        no        : Do not restrict learnt db size")
-		("dsched", custom(this, &SearchOptions::mapCustom), "Configure secondary deletion policy (see restarts)")
+		("dsched", storeTo(local->solve.reduce.sched, &SearchOptions::parseSchedule), "Configure secondary deletion policy (see restarts)")
 		("dglue", storeTo(solverOpts->reduceGlue), "Don't delete nogoods with lbd <= x")
 		("update-lbd", flag(solverOpts->updateLbd), "Dynamically update lbds of learnt nogoods")
 		("reduce-on-restart", flag(local->solve.reduce.reduceOnRestart), "Delete some learnt nogoods after every restart\n")
@@ -501,63 +505,59 @@ bool SearchOptions::mapInt(SearchOptions* this_, const std::string& n, const int
 	return false;
 }
 
+bool SearchOptions::parseSchedule(const std::string& s, ScheduleStrategy& sched) {
+	std::pair<std::string, std::vector<double> > v;
+	std::string temp = toLower(s);
+	bool        ok   = false;
+	if      (temp == "no")         { sched.init(0,0.0,0); return true;  }
+	else if (parse(temp, v.second)){ ok = true; }
+	else if (parse(temp, v))       { 
+		if      (v.first == "+" || v.first == "arith") { ok = v.second.size() >= 2; v.first = "+"; }
+		else if (v.first == "*" || v.first == "geom")  { ok = v.second.size() >= 2; }
+		else if (v.first == "l" || v.first == "luby")  { ok = v.second.size() >= 1; v.second.resize(1); }
+		else if (v.first == "b" || v.first == "biere") { ok = v.second.size() >= 3; v.second.resize(3); }
+	}
+	if (!ok) return false;
+	v.second.resize(3, 0.0);
+	sched.init(uint32(v.second[0]), v.second[1], (uint64)v.second[2], v.first == "+");
+	return true;
+}
+
 bool SearchOptions::mapCustom(SearchOptions* this_, const std::string& n, const std::string* value) {
-	if (n == "restarts" || n == "deletion" || n == "dsched") {
+	bool ok = false;
+	if (n == "deletion") {
 		std::vector<double> v;
-		ScheduleStrategy* sched = 0;
-		if (n != "deletion") {
-			sched = n[0] == 'r' 
-				? &this_->solveOpts->restart.sched 
-				: &this_->solveOpts->reduce.sched;
-		}
 		if (parse(*value, v)) {
-			if (sched) {
-				v.resize(3, 0.0);
-				sched->init(uint32(v[0]), v[1], uint32(v[2]));
-			}
-			else {
-				if (v.empty())  v.push_back(3.0);
-				if (v.size()==1)v.push_back(1.1);
-				if (v.size()==2)v.push_back(3.0);
-				this_->solveOpts->reduce.setStrategy(v[0], v[1], v[2]);
-			}
-			return true;
+			if (v.empty())  v.push_back(3.0);
+			if (v.size()==1)v.push_back(1.1);
+			if (v.size()==2)v.push_back(3.0);
+			this_->solveOpts->reduce.setStrategy(v[0], v[1], v[2]);
+			ok = true;
 		}
-		else if (*value == "no") {
-			if   (sched) { sched->init(0,0,0); }
-			else         { this_->solveOpts->reduce.disable = true;     }
-		}
-		else { return false; }
-		return true;
+		else if (*value == "no") { ok = (this_->solveOpts->reduce.disable = true); }
 	}
 	else if (n == "shuffle") {
 		std::pair<int, int> shuf;
-		if (parse(*value, shuf)) {
-			this_->solveOpts->setShuffleParams(shuf.first, shuf.second);
-			return true;
-		}
+		ok = parse(*value, shuf);
+		this_->solveOpts->setShuffleParams(shuf.first, shuf.second);
 	}
 	else if (n == "dinit") {
 		std::pair<int, int> init(0, -1);
-		if (parse(*value, init)) {
-			this_->solveOpts->reduce.setInit((uint32)init.first, (uint32)init.second);
-			return true;
-		}
+		ok = parse(*value, init);
+		this_->solveOpts->reduce.setInit((uint32)init.first, (uint32)init.second);
 	}
 	else if (n == "dfrac") {
 		double f = .75;
-		if (parse(*value, f)) {
-			this_->solveOpts->reduce.setReduceFraction(f);
-			return true;
-		}
+		ok       = parse(*value, f);
+		this_->solveOpts->reduce.setReduceFraction(f);
 	}
-	return false;
+	return ok;
 }
 
 bool SearchOptions::mapLimits(SearchOptions* this_, const std::string&, const std::string* value) {
 	std::pair<int, int> p(-1,-1);
 	if (parse(*value, p)) {
-		this_->local->solve.limits = SearchLimits(static_cast<uint64>(p.first), static_cast<uint64>(p.second));
+		this_->local->solve.limits = SolveLimits(static_cast<uint64>(p.first), static_cast<uint64>(p.second));
 		return true;
 	}
 	return false;
@@ -597,18 +597,27 @@ void ClaspOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOptio
 
 bool ClaspOptions::validateOptions(const ProgramOptions::ParsedOptions& vm, Messages& m) {
 	if (mode.validateOptions(vm, m) && search.validateOptions(vm, m)) {
-		if (config->numThreads() > 1) {
-			if (config->thread.portfolio.empty()) {
-				config->thread.forceGP = true;
-				for (uint32 i = 1; i < config->numThreads(); ++i) {
-					config->threadConfig(i)->initFrom(*config->threadConfig(0));
-				}
+		const char* port = 0; std::string mem;
+		if (!config->thread.portfolio.empty()) {
+			if (config->thread.portfolio == "default") {
+				port = portfolio_g;
 			}
 			else {
-				std::ifstream portfolio(config->thread.portfolio.c_str());
-				populateThreadConfigs(portfolio, vm, m);
+				std::ifstream test(config->thread.portfolio.c_str());
+				if (!test) {
+					m.error = "Could not open portfolio file '";
+					m.error += config->thread.portfolio;
+					m.error += "'";
+					return false;
+				}
+				if (!parsePortfolio(test, config->numThreads(), mem)) {
+					m.error = "Portfolio file has unrecognized format!";
+					return false;
+				}
+				port = mem.data();
 			}
 		}
+		populateThreadConfigs(port, vm, m);
 		return config->validate(m.error);
 	}
 	return false;
@@ -625,52 +634,57 @@ void ClaspOptions::applyDefaults(Input::Format f) {
 	}
 }
 
-bool ClaspOptions::populateThreadConfigs(std::istream& in, const ProgramOptions::ParsedOptions& vm, Messages& m) {
-	StringSeq configs;
-	for (std::string line; std::getline(in, line); ) {
+bool ClaspOptions::parsePortfolio(std::istream& in, uint32 max, std::string& mem) const {
+	mem.reserve(128);
+	for (std::string line; std::getline(in, line) && max; ) {
 		if (line.empty()) continue;
 		if (line[0] == '[') {
-			configs.push_back(line);
-			if ((uint32)configs.size() == config->numThreads()) {
-				break;
-			}
+			mem += line;
+			mem += '\0';
+			--max;
 		}
-		else if (line[0] != '#') {
-			m.error = "Portfolio file has unrecognized format!";
-			return false;
-		}
-	}		
-	if (configs.empty()) {
-		configs.push_back("");
+		else if (line[0] != '#') { return false; }
 	}
-	bool addSeed = false;
-	for (uint32 x = 0, i = 0; i != config->numThreads(); ++i) {
-		if (!parseConfig(configs[x], i, vm, m)) {
-			return false;
+	mem += '\0';
+	return true;
+}
+
+bool ClaspOptions::populateThreadConfigs(const char* port, const ProgramOptions::ParsedOptions& vm, Messages& m) {
+	if (port == 0) {
+		config->thread.forceGP = true;
+		for (uint32 i = 1; i < config->numThreads(); ++i) {
+			config->threadConfig(i)->initFrom(*config->threadConfig(0));
 		}
-		if (addSeed) {
-			config->threadConfig(i)->solver().strategies().rng.srand(i);
+	}
+	else {
+		bool addSeed = false;
+		const char* p= port;
+		for (uint32 i = 0; i != config->numThreads(); ++i) {
+			if (!parseConfig(p, i, vm, m)) { return false; }
+			if (addSeed)                   { config->threadConfig(i)->solver().strategies().rng.srand(i); }
+			if (*p)                        { p += strlen(p) + 1; }
+			if (!*p)                       { p  = port; addSeed = true; }
 		}
-		if (++x == (uint32)configs.size()) {
-			x = 0;
-			addSeed = true;
-		}		
 	}
 	return true;
 }
 
-bool ClaspOptions::parseConfig(const std::string& in, uint32 id, const ProgramOptions::ParsedOptions& vm, Messages& m) {
-	std::string::size_type pos = in.find("]:");
-	std::string configOpts, configName;
-	if (pos != std::string::npos) {
-		configName.assign(in.begin()+1, in.begin()+pos);
-		if ( (pos = in.find_first_not_of(' ', pos+2)) == std::string::npos ) {
-			pos = in.size();
+bool ClaspOptions::parseConfig(const char* in, uint32 id, const ProgramOptions::ParsedOptions& vm, Messages& m) {
+	assert(in);
+	std::string configOpts;
+	const char* nameBeg = in;
+	const char* nameEnd = in;
+	if (*nameBeg == '[') {
+		for (nameEnd = ++nameBeg; *nameEnd && *nameEnd != ']';) {
+			++nameEnd;
 		}
-		configOpts.assign(in.begin()+pos, in.end());
+	}
+	if (nameEnd[0] != ']' || nameEnd[1] != ':') {
+		nameEnd    = nameBeg;
+		configOpts = in;
 	}
 	else {
-		configOpts = in;
+		configOpts = nameEnd+2;
 	}
 	try {
 		ProgramOptions::OptionContext root;
@@ -682,6 +696,7 @@ bool ClaspOptions::parseConfig(const std::string& in, uint32 id, const ProgramOp
 		opts.initOptions(root, hidden);
 		root.add(hidden);
 		// 1. all from command-line
+		if (id != 0) { opts.local->initFrom(*config->master()); }
 		ProgramOptions::ParsedOptions parsed(vm);
 		// 2. all from current config
 		parsed.assign(ProgramOptions::parseCommandString(configOpts, root));
@@ -693,11 +708,25 @@ bool ClaspOptions::parseConfig(const std::string& in, uint32 id, const ProgramOp
 	}
 	catch(const std::exception& e) {
 		std::string err("Config '");
-		err += configName;
+		err.append(nameBeg, (nameEnd-nameBeg));
 		err += "': ";
 		err += e.what();
 		throw std::runtime_error(err);
 	}
 	return true;
 }
+
+const char* portfolio_g = {
+	     /* 0 */"[DEFAULT]: "
+	 "\0"/* 1 */"[CRAFTED]: --berk-max=512 --berk-once --otfs=1 --recursive-str --dinit=800,10000 --dsched=20000,1.1 --reverse-arcs=2 --opt-heu=1"
+	 "\0"/* 2 */"[INDUST]:  --berk-max=512 --berk-once --restarts=256 --reverse-arcs=2 --recursive-str --dinit=800,10000 --dsched=10000,1.1 --dscore=2 --update-lbd --dalgo=2 --opt-heu=2"
+	 "\0"/* 3 */"[LUBY-SP]: --heu=VSIDS --restarts=128 --save-p --otfs=1 --rand-w=false --contr=0 --opt-heu=3"
+	 "\0"/* 4 */"[STRONG]:  --heu=VSIDS --otfs=2 --reverse-arcs=2 --recursive-str --rand-w=false"
+	 "\0"/* 5 */"[LOCAL-R]: --berk-max=512 --restarts=100,1.5,1000 --local-restarts --rand-w=false --contr=0"
+	 "\0"/* 6 */"[SLOW]:    --berk-max=512 --restarts=16000 --initial-look=50"
+	 "\0"/* 7 */"[RANDY]:   --heu=VSIDS --rand-f=0.02 --rand-p=20,100 --recursive-str --shuffle=10,2"
+	 "\0"/* 8 */"[SIMPLE]:  --heu=VMTF --str=no --contr=0 --restarts=100,1.3 --dinit=800,10000"
+	 ""
+};
+
 }
