@@ -259,9 +259,11 @@ void DefaultOutput::reportProgress(Enumerator::ProgressType t, const Solver& s,
 
 void DefaultOutput::reportState(int state, bool enter, double time) {
 	if (enter && verbosity()) {
-		if (state == ClaspFacade::state_read) {
+		if (state == ClaspFacade::state_start) {
 			if (!solver().empty()) comment(1, "%s\n", solver().c_str());
 			if (!input().empty())  comment(1, "Reading from %s\n", prettify(input()).c_str());
+		}
+		else if (state == ClaspFacade::state_read) {
 			comment(2, "Reading      : ");
 		}
 		else if (state == ClaspFacade::state_preprocess) {
@@ -474,7 +476,11 @@ void DefaultOutput::printLemmaStats(const SolveStats& st) const {
 	printKeyValue("  Loop", "%-6"PRIu64, st.learnts[1]);
 	printf(" (Average Length: %.1f) \n", average(st.lits[1], st.learnts[1]));
 	printKeyValue("  Other", "%-6"PRIu64, st.learnts[2]);
-	printf(" (Average Length: %.1f) \n", average(st.lits[2], st.learnts[2]));
+	printf(" (Average Length: %.1f"     , average(st.lits[2], st.learnts[2]));
+	if (st.parallel != 0) {  
+		printf(" Unit: %"PRIu64" Average Jumps: %.2f", st.parallel->imps, average(st.parallel->jumps, st.parallel->imps));
+	}
+	printf(")\n");
 	printKeyValue("  Deleted", "%-6"PRIu64"\n", st.deleted);
 }
 
@@ -606,7 +612,7 @@ void JsonOutput::reportState(int state, bool enter, double) {
 		open_ = "";
 		startObject();
 	}
-	if (state == ClaspFacade::state_read && enter) {
+	if (state == ClaspFacade::state_start && enter) {
 		printKeyValue("Solver", solver().c_str());
 		printKeyValue("Input", input().c_str());
 	}
