@@ -26,7 +26,7 @@
 #endif
 
 #if !defined(CLASP_VERSION)
-#define CLASP_VERSION "2.0.2"
+#define CLASP_VERSION "2.0.3"
 #endif
 #if !defined(CLASP_LEGAL)
 #define CLASP_LEGAL \
@@ -124,10 +124,10 @@ struct ThreadOptions {
 		GRestarts():maxR(0) {}
 		uint32           maxR;
 		ScheduleStrategy sched;
-	} restarts;           /**< global restart strategy */
-	std::string portfolio;/**< portfolio file */
-	bool   forceGP;        /**< force guiding path scheme */
-	bool   genTemplate;    /**< generate a template portfolio file and exit */
+	} restarts;             /**< global restart strategy */
+	std::string portfolio;  /**< portfolio file */
+	bool        forceGP;    /**< force guiding path scheme */
+	bool        genTemplate;/**< generate a template portfolio file and exit */
 };
 
 
@@ -137,35 +137,33 @@ public:
 	typedef ProgramBuilder::EqOptions        EqOptions;
 	typedef Enumerator::ProgressReport       Progress;
 	typedef std::auto_ptr<Enumerator> EnumPtr;
+	enum EnumMode { enum_auto = 0, enum_bt = 1, enum_record = 2, enum_consequences = 4, enum_brave = 5, enum_cautious = 6 };
 	GlobalOptions();
 	Enumerator*     initEnumerator(Enumerator::Report* r = 0);
-	bool consequences() const { return enumerate.brave || enumerate.cautious; }
-	const char* cbType()const { return consequences() ? (enumerate.brave ? "Brave" : "Cautious") : "none"; }
+	bool consequences() const { return enumerate.consequences(); }
+	const char* cbType()const { return consequences() ? (enumerate.mode == enum_brave  ? "Brave" : "Cautious") : "none"; }
 	SharedContext ctx;    /**< Context-object used by all solvers */
 	ThreadOptions thread; /**< Options for parallel solving */
 	EqOptions     eq;     /**< options for equivalence preprocessing */
 	struct Optimize {     /**< Optimization options */
-		Optimize() : hierarch(0), heu(0), no(false), all(false), sat(false) {}
+		Optimize() : hierarch(0), no(false), all(false), sat(false) {}
 		WeightVec vals;     /**< initial values for optimize statements */
 		int    hierarch;    /**< use hierarchical optimization scheme */
-		int    heu;         /**< consider optimize statements in heuristics */
 		bool   no;          /**< ignore optimize statements */
 		bool   all;         /**< compute all models <= vals */
 		bool   sat;         /**< force max sat */
 	}    opt;
 	struct EnumOptions {  /**< Enumeration options */
-		EnumOptions() : progress(0), numModels(-1), projectOpts(7), project(false)
-		              , record(false), restartOnModel(false), brave(false)
-		              , cautious(false), onlyPre(false)  {}
-		Progress* progress; /**< enable progress reporting? */
-		int  numModels;     /**< number of models to compute */
-		int  projectOpts;   /**< options for projection */
-		bool project;       /**< enable projection */
-		bool record;        /**< enable solution recording */
-		bool restartOnModel;/**< restart after each model */
-		bool brave;         /**< enable brave reasoning */
-		bool cautious;      /**< enable cautious reasoning */
-		bool onlyPre;       /**< stop after preprocessing step? */
+		EnumOptions() : progress(0), numModels(-1), mode(enum_auto), projectOpts(7), project(false)
+		              , restartOnModel(false), onlyPre(false)  {}
+		bool consequences() const { return (mode & enum_consequences) != 0; }
+		Progress* progress;      /**< enable progress reporting? */
+		int       numModels;     /**< number of models to compute */
+		EnumMode  mode;          /**< enumeration type to use */
+		int       projectOpts;   /**< options for projection */
+		bool      project;       /**< enable projection */
+		bool      restartOnModel;/**< restart after each model */
+		bool      onlyPre;       /**< stop after preprocessing step? */
 	} enumerate;
 };
 
@@ -216,7 +214,7 @@ class ClaspFacade : public Enumerator::Report {
 public:
 	//! Defines the possible solving states
 	enum State { 
-		state_not_started,
+		state_start,       /*!< computation started */
 		state_read,        /*!< problem is read from input */
 		state_preprocess,  /*!< problem is prepared */
 		state_solve,       /*!< search is active */
@@ -320,9 +318,10 @@ private:
 	ClaspFacade(const ClaspFacade&);
 	ClaspFacade& operator=(const ClaspFacade&);
 	struct AutoState {
-		AutoState(ClaspFacade* f, State s) : self_(f) { self_->setState(s, event_state_enter); }
-		~AutoState() { self_->setState(self_->state(), event_state_exit); delete self_->ctrl_; self_->ctrl_ = 0; }
+		AutoState(ClaspFacade* f, State s) : self_(f), state_(s) { self_->setState(s, event_state_enter); }
+		~AutoState() { self_->setState(state_, event_state_exit); delete self_->ctrl_; self_->ctrl_ = 0; }
 		ClaspFacade* self_;
+		State        state_;
 	};
 	typedef SingleOwnerPtr<ProgramBuilder> Api;
 	typedef SingleOwnerPtr<SharedDependencyGraph> GraphPtr;
@@ -350,7 +349,6 @@ private:
 	bool   read();
 	bool   preprocess();
 	void   setProblemSize() const;
-	bool   configureMinimize(SharedMinimizeData* min) const;
 	bool   initEnumeration(SharedMinimizeData* min);
 	void   setGraph();
 	void   initSolveObject(ClaspConfig& config);
