@@ -73,8 +73,9 @@ public:
 	void   setIntegrate(uint32 grace, uint8 filter);
 	//! configures "global" restarts (only used if search-space splitting is active)
 	void   setRestarts(uint32 maxR, const ScheduleStrategy& rs);
-	//! must be called once for each solver except the master
+	//! attaches a new thread to s and runs it in this algorithm
 	/*!
+	 * shall be called once for each solver except the master
 	 * \pre s.id() != ParallelSolve::invalidId
 	 */
 	void   addThread(Solver& s, SolveParams& p);
@@ -86,27 +87,12 @@ public:
 	bool   integrateUseHeuristic() const { return intHeuristic_; }
 	uint32 integrateGrace()        const { return intGrace_; }
 	uint32 integrateFlags()        const { return intFlags_; }
-	//! terminates current solving process
+	//! terminates current solving process and all client threads
 	bool   terminate();
+	//! requests a global restart
 	void   requestRestart();
-	// interface for controlling the solve process
-	struct Message_t {
-		enum Type {
-			  /* control */ quit_message = 0, terminate_message = 1, unsat_message = 2, restart_message = 3
-			, /* other */   not_a_control_message = 4, split_message = 8, update_message = 16
-		};
-		struct Local {
-			Local() : any(0), update(0), doUpdate(false) {}
-			uint32 any;
-			uint32 update;
-			bool   doUpdate;
-		};
-	};
-	typedef Message_t::Type MessageType;	
-	bool   handleMessages(Solver& s, Message_t::Local& mq);
-	bool   checkUpdates(Message_t::Local& mq) const;
+	bool   handleMessages(Solver& s);
 	void   pushWork(LitVec& gp);
-	void   postMessage(MessageType);
 private:
 	ParallelSolve(const ParallelSolve&);
 	ParallelSolve& operator=(const ParallelSolve&);
@@ -126,45 +112,21 @@ private:
 	void   initQueue();
 	bool   requestWork(Solver& s, PathPtr& out);
 	bool   backtrackFromModel(Solver& s);
-	void   terminateUnsat(Solver& s, bool complete);
-	void   terminateComplete(bool complete);
-	bool   waitOnRestart(Solver& s);
+	void   terminate(Solver& s, bool complete);
+	bool   waitOnSync(Solver& s);
 	void   exception(Solver& s, PathPtr& path, ErrorCode e, const char* what);
 	// -------------------------------------------------------------------------------------------
 	struct SharedData;
-	SharedData*         shared_;      // Shared control data
-	ParallelHandler**   thread_;      // Thread-locl control data
-	struct ThreadControlFlags {
-	  ThreadControlFlags();
-		enum Flag { 
-			  quit_flag = 1, terminate_flag = 2, unsat_flag = 4, restart_flag = 8
-			, ctrl_message = quit_flag|terminate_flag|unsat_flag|restart_flag
-			, inhibit_restart_flag = 16, complete_flag = 32, allow_gp_flag = 64 
-		};
-		void reset();
-		bool quit()         const { return (control & uint32(quit_flag)) != 0; }
-		bool terminate()    const { return (control & uint32(quit_flag|terminate_flag)) != 0; }
-		bool ok()           const { return (control & uint32(terminate_flag)) != 0; }
-		bool restart()      const { return (control & uint32(unsat_flag|restart_flag))  != 0; }
-		bool unsatRestart() const { return (control & uint32(unsat_flag)) != 0; }
-		bool hasControlMsg()const { return (control & uint32(ctrl_message)) != 0; }
-		bool complete()     const { return (control & uint32(complete_flag)) != 0; }
-		bool allowRestart() const { return (control & uint32(inhibit_restart_flag)) == 0; }
-		bool allowSplit()   const { return (control & uint32(allow_gp_flag)) != 0; }
-		bool setFlag(Flag flag);
-		std::atomic<int>    workReq;   // > 0: someone needs work
-		std::atomic<uint32> restartReq;// == numThreads(): restart
-		std::atomic<uint32> control;   // set of active control flags
-		std::atomic<uint32> messages;  // global message counter
-		std::atomic<uint32> updates;   // global number of "update"  messages
-	}                    msg_;
-	ScheduleStrategy     globalR_;   // global restart strategy
-	uint32               maxThreads_;   // number of threads alloacated 
-	uint32               maxRestarts_;  // disable global restarts once reached 
-	uint32               intGrace_;     // grace period for clauses to integrate
-	uint32               intFlags_;     // bitset controlling clause integration
-	bool                 intHeuristic_; // use heuristic in clause integration
-	bool                 forceGP_;	    // force guiding path mode even if portfolio is used
+	// SHARED DATA
+	SharedData*       shared_;       // Shared control data
+	ParallelHandler** thread_;       // Thread-locl control data
+	// READ ONLY
+	uint32            maxThreads_;   // number of threads alloacated 
+	uint32            maxRestarts_;  // disable global restarts once reached 
+	uint32            intGrace_;     // grace period for clauses to integrate
+	uint32            intFlags_;     // bitset controlling clause integration
+	bool              intHeuristic_; // use heuristic in clause integration
+	bool              forceGP_;	     // force guiding path mode even if portfolio is used
 };
 
 
@@ -226,12 +188,11 @@ public:
 	bool isModel(Solver& s);
 
 	// own interface
+	
 	// TODO: make functions virtual once necessary 
 	
 	//! returns true if handler's guiding path is disjoint from all others
 	bool disjointPath() const { return gp_.split; }
-	//! returns true if the handler can currently split off a new guiding path
-	bool splittable()   const;
 	
 	//! called before solver starts to solve given guiding path
 	/*!
@@ -268,15 +229,12 @@ public:
 	 */
 	bool handleRestartMessage();
 
-	//! mark all update messages <= up as handled
-	void setUpdate(uint32 up) { messageHandler_.msg.update = up; }
-
 	SolveStats aggStats;  // aggregated statistics over all gps
 	//@}  
 private:
 	static void threadMain(ParallelHandler* h);
 	bool simplify(Solver& s, bool re);
-	bool integrateClauses(Solver& s, SharedLiterals** arr, uint32 num);
+	bool integrateClauses(Solver& s);
 	void add(ClauseHead* h);
 	ParallelSolve* ctrl() const { return messageHandler_.ctrl; }
 	typedef LitVec::size_type size_type;
@@ -302,15 +260,11 @@ private:
 		}
 	} gp_;
 	struct MessageHandler : PostPropagator {
-		typedef ParallelSolve::Message_t::Local MessageQ;
 		explicit MessageHandler(ParallelSolve* c) : ctrl(c) {}
 		uint32 priority() const { return PostPropagator::priority_highest; }
-		bool   propagateFixpoint(Solver& s) { return ctrl->handleMessages(s, msg); }
+		bool   propagateFixpoint(Solver& s) { return ctrl->handleMessages(s); }
 		bool   propagate(Solver& s)         { return MessageHandler::propagateFixpoint(s); }
-		bool   pendingUpdate() const        { return msg.doUpdate; }
-		bool   doUpdate(Solver& s, const GP& gp);
 		ParallelSolve* ctrl; // get messages from here
-		MessageQ       msg;  // local messages
 	}    messageHandler_;
 };
 
