@@ -34,9 +34,6 @@ namespace Clasp {
 // classes used by the Solver
 ///////////////////////////////////////////////////////////////////////////////
 class   DecisionHeuristic;
-class   PostPropagator;
-class   SharedContext;
-class   SatPreprocessor;
 
 /**
  * \defgroup solver Solver and related classes
@@ -68,9 +65,14 @@ struct SolverStrategies {
 		score_lbd,
 		score_both
 	};
+	enum HeuOpts {
+		heu_use_static_sign = 1,
+		heu_use_model       = 2
+	};
 	Heuristic           heuristic;
 	SearchStrategy      search;
 	RNG                 rng;
+	uint32              heuOpts;             /*!< Set of HeuOpts */
 	int                 saveProgress;
 	int                 reverseArcs;
 	int                 otfs;
@@ -119,6 +121,18 @@ private:
 	uint32 compress_;
 };
 
+struct SearchLimits {
+	explicit SearchLimits(uint64 conf = UINT64_MAX) 
+		: conflicts(conf)
+		, local(UINT64_MAX)
+		, learnts(UINT32_MAX) {
+	}
+	bool   reached() const { return conflicts == 0 || local == 0; }
+	uint64 conflicts; /*!< search for at least number of conflicts */
+	uint64 local;     /*!< search for at least number of conflicts in branch  */
+	uint32 learnts;   /*!< limit on number of learnt lemmas        */
+};
+
 //! clasp's Solver class
 /*!
  * A Solver-object maintains the state and provides the functions 
@@ -219,31 +233,18 @@ public:
 	uint32 numConstraints() const;
 
 	//@}
-	struct RestartLimit {
-		RestartLimit(uint64 c = UINT64_MAX, bool loc = false) : maxConf(c), countLocal(loc) {}
-		uint64 maxConf;
-		bool   countLocal;
-		bool   reached() const { return maxConf == 0; }
-	};
-	struct LearntLimit {
-		LearntLimit(uint64 c = UINT64_MAX, uint32 n = UINT32_MAX) : maxConf(c), maxLearnt(n) {}
-		uint64 maxConf;
-		uint32 maxLearnt;
-		bool   reached() const { return maxConf == 0; }
-	};
 	/*!
 	 * \name DPLL-functions
 	 */
 	//@{
-	//! searches for a model as long as none of the given limits is reached.
+	//! searches for a model as long as the given limit is not reached.
 	/*!
 	 * The search function implements clasp's CDNL-algorithm.
-	 * It searches for a model as long as neither of the limits given by r and d
-	 * is reached. The limits r and d are updated during search.
+	 * It searches for a model as long as neither of the limits given by limit
+	 * is reached. The limits are updated during search.
 	 *
 	 * \pre endInit() returned true and !hasConflict()
-	 * \param r restart limit  (stop after r.maxConf global or local conflicts)
-	 * \param d deletion limit (stop after d.maxConf or if numLearnts() > d.maxLearnt)
+	 * \param limit imposed limit on conflicts and number of learnt constraints
 	 * \param randProb pick next decision variable randomly with a probability of randProp
 	 * \return
 	 *  - value_true: if a model was found.
@@ -254,11 +255,13 @@ public:
 	 * \note search treats the root level as top-level, i.e. it
 	 * will never backtrack below that level.
 	 */
-	ValueRep search(RestartLimit& r, LearntLimit& d, double randProb = 0.0);
+	ValueRep search(SearchLimits& limit, double randProb = 0.0);
 	ValueRep search(uint64 maxC, uint32 maxL, bool local = false, double rp  = 0.0) {
-		RestartLimit r(maxC, local);
-		LearntLimit  d(UINT64_MAX, maxL);
-		return search(r, d, rp);
+		SearchLimits limit;
+		if (!local) { limit.conflicts = maxC; }
+		else        { limit.conflicts = UINT64_MAX; limit.local = maxC; }
+		limit.learnts = maxL;
+		return search(limit, rp);
 	}
 
 	//! moves the root-level i levels down (i.e. away from the top-level)
@@ -328,7 +331,7 @@ public:
 	/*!
 	 * \note if this method is called on a decision-level > 0 it is a noop and will
 	 * simply return true.
-	 * \return fallse, if a top-level conflict is detected. Otherwise, true.
+	 * \return false, if a top-level conflict is detected. Otherwise, true.
 	 */
 	bool simplify();
 
@@ -412,15 +415,14 @@ public:
 	//! selects and assumes the next branching literal by calling the installed decision heuristic.
 	/*!
 	 * \pre queueSize() == 0
-	 * \note the next decision literal will be selected randomly with a
-	 * probability set using the initRandomHeuristic method.
+	 * \note the next decision literal will be selected randomly with probability f.
 	 * \return 
 	 *  - true if the assignment is not total and a literal was assumed (or forced).
 	 *  - false otherwise
 	 *  .
 	 * \see DecisionHeuristic
 	 */
-	bool decideNextBranch();
+	bool decideNextBranch(double f = 0.0);
 
 	//! Sets a conflict that forces the solver to terminate its search
 	/*!
@@ -462,13 +464,13 @@ public:
 	 * \param c the constraint that wants to test p (may be 0)
 	 * \pre p is free
 	 * \note If c is not null and testing p does not lead to a conflict, 
-	         c->undoLevel() is called *before* p is undone. Hence, the
-			 range [s.levelStart(s.decisionLevel()), s.assignment().size())
-			 contains p followed by all literals that were forced because of p.
+	 *       c->undoLevel() is called *before* p is undone. Hence, the
+	 *       range [s.levelStart(s.decisionLevel()), s.assignment().size())
+	 *       contains p followed by all literals that were forced because of p.
 	 * \note During propagation of p, only post propagators with priority
-	 * in the range [priority_highest, priority_lookahead) are called.
+	 *       higher than that of c are called.
 	 */
-	bool test(Literal p, Constraint* c);
+	bool test(Literal p, PostPropagator* c);
 
 	//! estimates the number of assignments following from setting p to true.
 	/*!
@@ -526,11 +528,9 @@ public:
 	 * The function checks for symmetric models, i.e. models that differ only in the 
 	 * assignment of variables outside of the solver's assignment. 
 	 * Typical example: vars eliminated by the SAT-preprocessor
-	 * \param expand if false, any symmetric models are ignored. Otherwise, symmetric models
-	 *        are expanded and stored in the solver.
 	 * \pre the current assignment is a model
 	 */ 
-	bool nextSymModel(bool expand);
+	bool nextSymModel();
 
 	//@}  
 
@@ -872,11 +872,13 @@ private:
 		void simplify(Solver& s, bool shuf);
 		void reset();
 		bool isModel(Solver& s);
-		bool nextSymModel(Solver& s, bool expand);
-		bool init(Solver& s);
+		bool init(Solver& s, PostPropagator* list);
+		bool init(Solver& s) { return init(s, head); }
+		PostPropagator* enableFunky();
+		PostPropagator* disableFunky();
+		PostPropagator* nil() const { return (PostPropagator*)1; }
 		PostPropagator* head;
 		PostPropagator* look;
-		PostPropagator* saved;
 	};
 	struct CmpScore {
 		CmpScore(const ConstraintDB& learnts, const SolverStrategies& st) : db(learnts), strat(st) {}
@@ -896,7 +898,6 @@ private:
 	inline  bool hasStopConflict() const;
 	bool    validWatch(Literal p) const { return p.index() < (uint32)watches_.size(); }
 	uint32  mark(uint32 s, uint32 e);
-	void    initRandomHeuristic(double randFreq);
 	void    freeMem();
 	bool    simplifySAT();
 	void    simplifyShort(Literal p);
@@ -932,7 +933,6 @@ private:
 	CCMinRecursive*   ccMin_;       // additional data for supporting recursive strengthen
 	SmallClauseAlloc* smallAlloc_;  // allocator object for small clauses
 	SharedContext*    shared_;      // initialized by master thread - otherwise read-only!
-	DecisionHeuristic*randHeuristic_;
 	VarVec*           levConflicts_;// For a DL d, levConflicts_[d-1] stores num conflicts when d was started
 	ConstraintDB*     undoHead_;    // Free list of undo DBs
 	Constraint*       enum_;        // enumeration constraint - set by enumerator
@@ -950,7 +950,6 @@ inline bool isRevLit(const Solver& s, Literal p, uint32 maxL) {
 	return s.isFalse(p) && (s.seen(p) || s.level(p.var()) < maxL);
 }
 
-
 //@}
 
 /**
