@@ -378,7 +378,11 @@ struct ClauseCreator::Watches {
 	Watches() : fwLev(0), swLev(0) {
 		lits[0] = lits[1] = lits[2] = negLit(0);
 	}
-	void init(Solver& s, const Literal* it, const Literal* end) {
+	void init(Solver& s, const Literal* it, const Literal* end, bool checkOrder) {
+		if (!checkOrder) {
+			std::memcpy(lits, it, std::min(uint32(end - it), uint32(Clause::MAX_SHORT_LEN))*sizeof(Literal));
+			return;
+		}
 		for (uint32 pos = 0; it != end; ++it) {
 			Literal p      = *it;
 			uint32 abstr_p = Detail::levelAbstraction(s, p);
@@ -407,41 +411,40 @@ ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause
 	assert(!s.hasConflict() && "ClauseCreator::integrate() - precondition violated!");
 	Detail::Sink shared( (modeFlags & clause_no_release) == 0 ? clause : 0);
 	// determine state of clause
-	if (clause->size() > 1) {
-		Watches w;
-		if ( (modeFlags & clause_known_order) == 0 ) {
-			w.init(s, clause->begin(), clause->end());
+	Result  result;
+	Watches w;
+	w.init(s, clause->begin(), clause->end(), (modeFlags & clause_known_order) == 0);
+	Status  stat(status(s, w.lits, modeFlags));
+	if (clause->size() > 1 && stat != status_subsumed) {
+		if (clause->size() > Clause::MAX_SHORT_LEN) {
+			result.local  = Clause::newShared(s, clause, e, w.lits, shared.clause == 0);
+			shared.clause = 0;
+		}
+		else if (clause->size() > 3 || (modeFlags & clause_explicit) != 0) {
+			result.local  = Clause::newClause(s, w.lits, clause->size(), e);
 		}
 		else {
-			std::memcpy(w.lits, clause->begin(), std::min(clause->size(), uint32(Clause::MAX_SHORT_LEN))*sizeof(Literal));
+			// implicitly shared via binary/ternary implication graph;
+			// only check for implication/conflict but do not create
+			// a local representation for the clause
+			s.stats.addLearnt(clause->size(), e.type());
+			modeFlags    |= clause_no_add;
 		}
-		Status stat = status(s, w.lits, modeFlags);
-		if (stat != status_subsumed) {
-			Literal imp = stat == status_unit ? w.lits[0] : posLit(0);
-			ClauseHead* x = 0;
-			if (clause->size() > Clause::MAX_SHORT_LEN) {
-				x = Clause::newShared(s, clause, e, w.lits, shared.clause == 0);
-				shared.clause = 0;
-			}
-			else if (clause->size() > 3 || (modeFlags & clause_explicit) != 0) {
-				x = Clause::newClause(s, w.lits, clause->size(), e);
-			}
-			else {
-				// implicitly shared via binary/ternary implication graph;
-				// only check for implication/conflict but do not create
-				// a local representation for the clause
-				Antecedent ante(~w.lits[1], ~w.lits[2]);
-				return Result(x, imp == posLit(0) || s.force(imp, s.level(w.lits[1].var()), ante));
-			}
-			if ( (modeFlags & clause_no_add) == 0 ) { s.addLearnt(x, clause->size()); }
-			return Result(x, imp == posLit(0) || s.force(imp, s.level(w.lits[1].var()), x));
+		if ((modeFlags & clause_no_add) == 0) { s.addLearnt(result.local, clause->size()); }
+	}
+	if (stat == status_unit) {
+		Antecedent ante = result.local ? Antecedent(result.local) : Antecedent(~w.lits[1], ~w.lits[2]);
+		uint32 DL       = s.decisionLevel();
+		uint32 impLevel = s.level(w.lits[1].var());
+		result.ok       = impLevel
+		                ? s.force(w.lits[0], impLevel, ante) 
+		                :(s.stats.addLearnt(1, e.type()), s.addUnary(w.lits[0], Constraint_t::static_constraint));
+		if (s.stats.parallel) {
+			s.stats.parallel->imps  += 1;
+			s.stats.parallel->jumps += (DL - s.decisionLevel());
 		}
-		return Result(0, true);
-	}
-	else {
-		Literal imp = clause->size() > 0 ? *clause->begin() : negLit(0);
-		return Result(0, s.addUnary(imp, Constraint_t::static_constraint));
 	}
+	return result;
 }
 ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags) { 
 	return integrate(s, clause, modeFlags, ClauseInfo().setType(clause->type())); 
@@ -674,6 +677,7 @@ bool Clause::isSatisfied(const Solver& s, LitVec& freeLits) {
 	if (ClauseHead::satisfied(s)) {
 		return true;
 	}
+	assert(s.queueSize() == 0 && "Watches might be false!");
 	freeLits.push_back(head_[0]);
 	freeLits.push_back(head_[1]);
 	if (!s.isFalse(head_[2])) freeLits.push_back(head_[2]);
