@@ -145,30 +145,99 @@ private:
 /////////////////////////////////////////////////////////////////////////////////////////
 struct ParallelSolve::SharedData {
 	typedef tbb::concurrent_queue<const LitVec*> queue;
+	enum MsgFlag {
+		terminate_flag     = 1u, sync_flag  = 2u,  split_flag    = 4u, 
+		restart_flag       = 8u, unsat_flag = 16u, complete_flag = 32u,
+		interrupt_flag     = 64u,  // set on terminate from outside
+		allow_gp_flag      = 128u, // set if splitting mode is active
+		forbid_restart_flag= 256u, // set if restarts are no longer allowed
+		cancel_restart_flag= 512u  // set if last restart request was cancelled
+	};
+	enum Message {
+		msg_terminate      = (terminate_flag),
+		msg_interrupt      = (terminate_flag | interrupt_flag),
+		msg_sync_restart   = (sync_flag | restart_flag),
+		msg_sync_unsat     = (sync_flag | unsat_flag | restart_flag),
+		msg_split          = split_flag
+	};
 	SharedData() : path(0) { reset(0, 1); }
 	void reset(SharedContext* a_ctx, uint32 numT) {
 		clearQueue();
-		init.reset();
+		syncT.reset();
 		workSem.unsafe_init(0, numT);
-		ctx    = a_ctx;
-		path   = 0;
-		nextId = 1;
+		globalR.reset();
+		maxConflict = globalR.current();
+		ctx         = a_ctx;
+		path        = 0;
+		nextId      = 1;
+		workReq     = 0;
+		restartReq  = 0;
+		control     = 0;
 	}
 	void clearQueue() {
 		for (const LitVec* a = 0; workQ.try_pop(a); ) {
 			if (a != path) { delete a; }
 		}
 	}
-	Enumerator* enumerator() const { return ctx->enumerator(); }
-	SharedContext*    ctx;     // shared context object
-	const LitVec*     path;    // initial guiding path - typically empty
-	Timer<RealTime>   init;    // thread init time
-	tbb::mutex        modelM;  // model-mutex 
-	BarrierSemaphore  workSem; // work-semaphore
-	queue             workQ;   // work-queue
-	uint32            nextId;  // next solver id to use
+	Enumerator* enumerator()  const { return ctx->enumerator(); }
+	// MESSAGES
+	bool        hasMessage()  const { return (control & uint32(7)) != 0; }
+	bool        synchronize() const { return (control & uint32(sync_flag))      != 0; }
+	bool        terminate()   const { return (control & uint32(terminate_flag)) != 0; }
+	bool        split()       const { return (control & uint32(split_flag))     != 0; }
+	bool        postMessage(Message m, bool notify);
+	void        aboutToSplit()      { if (--workReq == 0) updateSplitFlag();  }
+	void        updateSplitFlag();
+	// CONTROL FLAGS
+	bool        hasControl(uint32 f) const { return (control & f) != 0;        }
+	bool        interrupt()          const { return hasControl(interrupt_flag);}
+	bool        complete()           const { return hasControl(complete_flag); }
+	bool        restart()            const { return hasControl(restart_flag);  }
+	bool        allowSplit()         const { return hasControl(allow_gp_flag); }
+	bool        allowRestart()       const { return !hasControl(forbid_restart_flag); }
+	bool        setControl(uint32 flags)   { return (fetch_and_or(control, flags) & flags) != flags; }
+	bool        clearControl(uint32 flags) { return (fetch_and_and(control, ~flags) & flags) == flags; }
+	ScheduleStrategy    globalR;     // global restart strategy
+	uint64              maxConflict; // current restart limit
+	SharedContext*      ctx;         // shared context object
+	const LitVec*       path;        // initial guiding path - typically empty
+	Timer<RealTime>     syncT;       // thread sync time
+	tbb::mutex          modelM;      // model-mutex 
+	BarrierSemaphore    workSem;     // work-semaphore
+	queue               workQ;       // work-queue
+	uint32              nextId;      // next solver id to use
+	std::atomic<int>    workReq;     // > 0: someone needs work
+	std::atomic<uint32> restartReq;  // == numThreads(): restart
+	std::atomic<uint32> control;     // set of active message flags
 };
 
+// post message to all threads
+bool ParallelSolve::SharedData::postMessage(Message m, bool notifyWaiting) {
+	if (m == msg_split) {
+		if (++workReq == 1) { updateSplitFlag(); }
+		return true;
+	}
+	else if (setControl(m)) {
+		// control message - notify all if requested
+		if (notifyWaiting) workSem.reset();
+		if ((uint32(m) & uint32(sync_flag)) != 0) {
+			syncT.reset();
+			syncT.start();
+		}
+		return true;
+	}
+	return false;
+}
+
+void ParallelSolve::SharedData::updateSplitFlag() {
+	for (bool splitF;;) {
+		splitF = (workReq > 0);	
+		if (split() == splitF) return;
+		if (splitF) fetch_and_or(control,   uint32(split_flag)); 
+		else        fetch_and_and(control, ~uint32(split_flag));
+	}
+}
+
 // Callable function object to be used as "main" function
 // for threads. Simply forwards to ParallelSolve::doSolve()
 struct ParallelSolve::EntryPoint {
@@ -179,34 +248,10 @@ struct ParallelSolve::EntryPoint {
 /////////////////////////////////////////////////////////////////////////////////////////
 // ParallelSolve
 /////////////////////////////////////////////////////////////////////////////////////////
-ParallelSolve::ThreadControlFlags::ThreadControlFlags() {
-	reset();
-}
-void ParallelSolve::ThreadControlFlags::reset() {
-	workReq    = 0;
-	restartReq = 0;
-	control    = 0;
-	messages   = 0;
-	updates    = 0;
-}
-
-// sets flag flag if not yet set
-// return: true if flag was set, false if flag is already set
-bool ParallelSolve::ThreadControlFlags::setFlag(Flag flag) {
-	for (uint32 x = control, y = flag; (x & y) == 0;) {
-		if (control.compare_and_swap(x|y, x) == x) {
-			return true;
-		}
-		x = control;
-	}
-	return false;
-}
-
 ParallelSolve::ParallelSolve()
 	: SolveAlgorithm()
 	, shared_(new SharedData)
 	, thread_(0)
-	, globalR_(0,0.0,0)
 	, maxThreads_(0)
 	, maxRestarts_(0)
 	, intGrace_(1024)
@@ -250,17 +295,16 @@ void ParallelSolve::setIntegrate(uint32 grace, uint8 filter) {
 }
 
 void ParallelSolve::setRestarts(uint32 maxR, const ScheduleStrategy& rs) {
-	maxRestarts_ = maxR;
-	globalR_     = rs;
+	shared_->globalR     = rs;
+	shared_->maxConflict = shared_->globalR.current();
+	maxRestarts_         = maxR;
 }
 
 // prepare parallel solving
 SharedContext* ParallelSolve::init(int numT, SharedContext& ctx) {
 	assert(numT >= 1 && "Illegal number of threads");
 	shared_->reset(&ctx, static_cast<uint32>(numT));
-	msg_.reset();
 	reserveThreads();
-	shared_->init.start();
 	return &ctx;
 }
 
@@ -298,8 +342,8 @@ void ParallelSolve::destroyThread(uint32 id) {
 
 // joins with and destroys all active threads
 void ParallelSolve::joinThreads() {
-	shared_->init.reset();
-	shared_->init.start();
+	shared_->syncT.reset();
+	shared_->syncT.start();
 	LOG(0, "STARTING  shutdown");
 	int error = thread_[0]->error();
 	for (uint32 i = 1, end = maxThreads_; i != end; ++i) {
@@ -308,10 +352,10 @@ void ParallelSolve::joinThreads() {
 		}
 		destroyThread(i);
 	}
-	thread_[0]->setError(msg_.ok() ? thread_[0]->error() : error);
+	thread_[0]->setError(!shared_->interrupt() ? thread_[0]->error() : error);
 	maxThreads_ = 1;
-	shared_->init.stop();
-	LOG(0, "COMPLETED shutdown after %.3fs", shared_->init.total());
+	shared_->syncT.stop();
+	LOG(0, "COMPLETED shutdown after %.3fs", shared_->syncT.total());
 }
 
 // attach new client thread
@@ -330,7 +374,7 @@ void ParallelSolve::addThread(Solver& s, SolveParams& p) {
 void ParallelSolve::startSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
 	uint32 id = s.id();
 	LOG(id, "ENTERING  algorithm");
-	if (!msg_.terminate()) {
+	if (!shared_->terminate()) {
 		if (shared_->ctx->master()->stats.jumps) {
 			s.stats.enableJumpStats();
 		}
@@ -340,46 +384,32 @@ void ParallelSolve::startSolve(Solver& s, const SolveParams& p, const LitVec& as
 			// already called.
 			thread_[0]->init(s);
 			shared_->path = &assume;
+			shared_->syncT.start();
 		}
-		else if (!shared_->ctx->attach(s) && !msg_.quit()) {
+		else if (!shared_->ctx->attach(s) && !shared_->interrupt()) {
 			// problem is UNSAT - terminate algorithm
 			LOG(s.id(), "TRIVIALLY UNSAT!");
-			terminateComplete(true);
+			terminate(s, true);
 			return;
 		}
-		// wait for other threads
-		if (shared_->workSem.wait()) {
-			assert(msg_.workReq == 0 && shared_->workQ.empty());
-			initQueue(); // add initial path to work queue
-			shared_->init.stop();
-			LOG(id, "COMPLETED initialization after %.3fs", shared_->init.total());
-			shared_->workSem.reset();
-		}
+		waitOnSync(s);
 	}
 }
 
 // main solve loop executed by all threads
 bool ParallelSolve::doSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
-	PathPtr a(0);
+	PathPtr a(0); InitParams init = p.init;
 	try {
 		startSolve(s, p, assume);
-		MinimizeConstraint* min = 0;
 		for (; requestWork(s, a);) {
-			thread_[s.id()]->prepareForGP(*a, globalR_.current(), a.is_owner());
+			thread_[s.id()]->prepareForGP(*a, shared_->maxConflict, a.is_owner());
 			s.stats.reset();
-			if (initPath(s, *a) && (!min || min->integrateNext(s))) {
+			if (initPath(s, *a, init)) {
 				if (solvePath(s, p) == value_free) {
-					terminateComplete(false);
-					s.stats.parallel->terminated = true;
+					terminate(s, false);
 				}
 				s.clearStopConflict();
 			}
-			if (min || shared_->enumerator()->optimizeHierarchical()) {
-				// path is unsat and hierarchical optimization is active;
-				// relax minimize constraint to avoid problems during simplification
-				min  = shared_->enumerator()->constraint(s)->minimize();
-				min->restoreOptimum();
-			}
 		}
 	}
 	catch (const std::bad_alloc&)      { exception(s,a,error_oom, "std::bad_alloc" ); }
@@ -392,19 +422,19 @@ bool ParallelSolve::doSolve(Solver& s, const SolveParams& p, const LitVec& assum
 // detaches the solver<->handler connection and joins with
 // all client threads
 bool ParallelSolve::endSolve(Solver& s) {
-	assert(msg_.terminate() || thread_[s.id()]->error() != error_none);
+	assert(shared_->terminate() || thread_[s.id()]->error() != error_none);
 	LOG(s.id(), "LEAVING   algorithm");
 	uint32 id    = s.id();
 	int    error = thread_[id]->error();
 	thread_[id]->detach();
-	if (!msg_.quit() && s.sharedContext() && error == error_none) {
+	if (!shared_->interrupt() && s.sharedContext() && error == error_none) {
 		s.clearAssumptions();
 		if (s.hasConflict()) {
-			msg_.setFlag(ThreadControlFlags::complete_flag);
+			shared_->setControl(SharedData::complete_flag);
 		}
 	}
 	// this thread is leaving
-	shared_->workSem.removeParty(msg_.terminate());
+	shared_->workSem.removeParty(shared_->terminate());
 	s.stats.reset();
 	if (error != error_none) {
 		// solver terminated because of an exception
@@ -430,9 +460,9 @@ bool ParallelSolve::endSolve(Solver& s) {
 				default:            throw std::runtime_error("UNKNOWN ERROR!");
 			}
 		}
-		shared_->enumerator()->reportResult(msg_.complete());
+		shared_->enumerator()->reportResult(shared_->complete());
 	}
-	return !msg_.complete();
+	return !shared_->complete();
 }
 
 void ParallelSolve::exception(Solver& s, PathPtr& path, ErrorCode e, const char* what) {
@@ -440,10 +470,10 @@ void ParallelSolve::exception(Solver& s, PathPtr& path, ErrorCode e, const char*
 		LOG(s.id(), "EXCEPTION %s", what); (void) what;
 		thread_[s.id()]->setError(e);
 		if (s.id() == 0 || shared_->workSem.active()) { 
-			terminate();
+			ParallelSolve::terminate();
 			return;
 		}
-		else if (msg_.allowSplit() && path.get()) {
+		else if (shared_->allowSplit() && path.get()) {
 			shared_->workQ.push(path.release());
 			shared_->workSem.up();
 		}
@@ -453,73 +483,24 @@ void ParallelSolve::exception(Solver& s, PathPtr& path, ErrorCode e, const char*
 
 // forced termination from outside
 bool ParallelSolve::terminate() {
-	// do not use postMessage() to avoid possible
-	// deadlock because of unblockAll()!
-	msg_.setFlag(ThreadControlFlags::quit_flag);
-	++msg_.messages;
+	// do not notify blocked threads to avoid possible
+	// deadlock in semaphore!
+	shared_->postMessage(SharedData::msg_interrupt, false);
+	// notify any thread currently in Enumerator::backtrackFromModel()
+	shared_->ctx->enumerator()->terminate();
 	return true;
 }
 
-// terminated from inside of algorithm
-void ParallelSolve::terminateComplete(bool complete) {
-	postMessage(Message_t::terminate_message);
-	if (complete) {
-		msg_.setFlag(ThreadControlFlags::complete_flag);
-	}
-}
-
-// terminated because of unsat, check if there is more to do
-void ParallelSolve::terminateUnsat(Solver& s, bool complete) {
-	if (!msg_.terminate()) {
-		if (complete && shared_->enumerator()->optimizeHierarchical() && s.popRootLevel(s.rootLevel(), false)) {
-			// Problem is unsat and hierarchical optimization is active.
-			// The active level is at its optimum, but lower-priority levels might
-			// still be non-optimal.
-			// Notify other threads to prepare for solving next priority level.
-			LOG(s.id(), "COMPLETED current optimization step"); 
-			postMessage(Message_t::unsat_message);
-			assert(s.decisionLevel() == 0);
-			return;
-		}
-		// nothing more to do - post terminate message
-		terminateComplete(complete);
-	}
-}
-
-// post message to all threads
-void ParallelSolve::postMessage(MessageType m) {
-	if (m < Message_t::not_a_control_message) {
-		// control message - notify all if new
-		ThreadControlFlags::Flag cf = static_cast<ThreadControlFlags::Flag>((uint32(1)<<m));
-		if (msg_.setFlag(cf)) {
-			++msg_.messages;
-			shared_->workSem.reset();
-		}
-	}
-	else if (m == Message_t::split_message) {
-		++msg_.workReq;
-		++msg_.messages;
-	}
-	else if (m == Message_t::update_message) {
-		++msg_.updates;
-		++msg_.messages;
-	}
-	else {
-		assert("ERROR: Unknown message type!\n");
-	}
-}
-
-
 // tries to get new work for the given solver
 bool ParallelSolve::requestWork(Solver& s, PathPtr& out) { 
 	const LitVec* a;
-	while (!msg_.terminate()) {
+	while (!shared_->terminate()) {
 		// only clear path and stop conflict - we don't propagate() here
 		// because we would then have to handle any eventual conflicts
 		if (!s.popRootLevel(s.rootLevel(), false)) {
 			// s has a real top-level conflict - 
 			// there is nothing we can do to continue
-			terminateComplete(true); 
+			terminate(s, true);
 		}	
 		else if (shared_->workQ.try_pop(a)) {
 			assert(s.decisionLevel() == 0);
@@ -533,66 +514,89 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 			// or a real conflict - we'll handle it in the next iteration
 			// via the call to popRootLevel()
 		}
-		else if (msg_.restart()) {
-			// a restart request is active - we are fine with
+		else if (shared_->synchronize()) {
+			// a synchronize request is active - we are fine with
 			// this but did not yet had a chance to react on it
-			waitOnRestart(s);
+			waitOnSync(s);
 		}
-		else if (msg_.allowSplit()) {
+		else if (shared_->allowSplit()) {
 			// gp mode is active - request a split	
 			// and wait until someone has work for us
-			postMessage(Message_t::split_message);
-			if (!shared_->workSem.down() && !msg_.restart()) {
+			shared_->postMessage(SharedData::msg_split, false);
+			if (!shared_->workSem.down() && !shared_->synchronize()) {
 				// we are the last man standing, there is no
 				// work left - quitting time?
-				terminateUnsat(s, true);
+				terminate(s, true);
 			}
 		}
 		else {
 			// portfolio mode is active - no splitting, no work left
 			// quitting time? 
-			terminateUnsat(s, true);
+			terminate(s, true);
 		}
 	}
 	return false;
 }
 
-// handles an active restart request
-// returns true to signal that s should restart otherwise false
-bool ParallelSolve::waitOnRestart(Solver& s) {
-	assert(msg_.restart());
-	// react once
-	if (!thread_[s.id()]->handleRestartMessage()) {
-		msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
+// terminated from inside of algorithm
+// check if there is more to do
+void ParallelSolve::terminate(Solver& s, bool complete) {
+	if (!shared_->terminate()) {
+		if (complete && shared_->enumerator()->optimizeHierarchical() && s.popRootLevel(s.rootLevel(), false)) {
+			// Problem is unsat and hierarchical optimization is active.
+			// The active level is at its optimum, but lower-priority levels might
+			// still be non-optimal.
+			// Notify other threads to prepare for solving next priority level.
+			assert(s.decisionLevel() == 0);
+			complete = false;
+			LOG(s.id(), "COMPLETED current optimization step"); 
+			shared_->postMessage(SharedData::msg_sync_unsat, true);
+		}
+		else {
+			shared_->postMessage(SharedData::msg_terminate, true);
+			if (complete) { shared_->setControl(SharedData::complete_flag); }
+		}
 	}
-	LOG(s.id(), "WAITING   for restart");
+	if (!complete) { s.stats.parallel->terminated = true; }
+}
+
+// handles an active synchronization request
+// returns true to signal that s should restart otherwise false
+bool ParallelSolve::waitOnSync(Solver& s) {
+	if (!thread_[s.id()]->handleRestartMessage()) {
+		shared_->setControl(SharedData::forbid_restart_flag);
+	}
+	LOG(s.id(), "WAITING   for synchronization");
 	if (shared_->workSem.wait()) {
-		// last man standing - complete restart request
-		msg_.workReq   = 0;
-		msg_.restartReq= 0;
-		bool hasUnsat  = msg_.unsatRestart();
-		msg_.control  -= (msg_.control & (uint32(ThreadControlFlags::unsat_flag|ThreadControlFlags::restart_flag)));
-		if (!hasUnsat) {
-			globalR_.next();
-			if ( (maxRestarts_ -= maxRestarts_ > 0) == 0 ) {
-				globalR_.init(0, 0.0, 0);
+		// last man standing - complete synchronization request
+		shared_->workReq     = 0;
+		shared_->restartReq  = 0;
+		bool unsat           = shared_->hasControl(SharedData::unsat_flag);
+		if (shared_->globalR.index() != maxRestarts_) {
+			shared_->maxConflict = shared_->globalR.current();
+			shared_->globalR.next();
+		}
+		else {
+			shared_->maxConflict = UINT64_MAX;
+		}
+		if (unsat || shared_->allowRestart()) {
+			initQueue(); // add initial path to work queue
+			shared_->clearControl(SharedData::cancel_restart_flag);
+			if (unsat && !shared_->enumerator()->optimizeNext()) {
+				// nothing more to do
+				shared_->setControl(SharedData::terminate_flag | SharedData::complete_flag);
 			}
 		}
-		if (hasUnsat || msg_.allowRestart()) {
-			initQueue();
+		else {
+			shared_->setControl(SharedData::cancel_restart_flag);
 		}
-		if (hasUnsat && shared_->enumerator()->optimizeHierarchical()) {
-			Enumerator*         en = shared_->enumerator();
-			MinimizeConstraint* m  = en->constraint(s)->minimize();
-			if (en->enumerated == 0 || !m->shared_unsafe()->optimizeNext()) {
-				terminateComplete(true);
-			}
-		}
-		LOG(s.id(), "COMPLETED restart");
+		shared_->clearControl(SharedData::sync_flag | SharedData::split_flag | SharedData::unsat_flag | SharedData::restart_flag);
+		shared_->syncT.stop();
+		LOG(s.id(), "COMPLETED synchronization after %.3fs", shared_->syncT.total());
 		// wake up all blocked threads
 		shared_->workSem.reset();
 	}
-	return msg_.terminate() || (!msg_.restart() && msg_.allowRestart());
+	return shared_->terminate() || !shared_->hasControl(SharedData::cancel_restart_flag);
 }
 
 // If guiding path scheme is active only one
@@ -605,11 +609,11 @@ void ParallelSolve::initQueue() {
 	shared_->clearQueue();
 	int end = numThreads();
 	if (forceGP_) {
-		msg_.setFlag(ThreadControlFlags::allow_gp_flag);
+		shared_->setControl(SharedData::allow_gp_flag);
 		end = 1;
 	}
 	else {
-		msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
+		shared_->setControl(SharedData::forbid_restart_flag);
 	}
 	for (int i = 0; i != end; ++i) {
 		shared_->workQ.push(shared_->path);
@@ -635,89 +639,72 @@ bool ParallelSolve::backtrackFromModel(Solver& s) {
 		// first check if the model is still valid once all
 		// information is integrated into the solver
 		uint32 dl = s.decisionLevel();
-		if (!thread_[s.id()]->isModel(s)) {
-			// model no longer a (unique) model - continue search
+		if (shared_->terminate() || !thread_[s.id()]->isModel(s)) {
+			// model no longer a (unique) model or enough models enumerated
 			s.stats.removeModel(dl);
-			return true;
+			return !shared_->terminate(); // continue search?
 		}
-		r = shared_->enumerator()->backtrackFromModel(s);
+		r = shared_->enumerator()->backtrackFromModel(s, false);
 		if (r == Enumerator::enumerate_stop_enough || (r == Enumerator::enumerate_stop_complete && s.decisionLevel() == 0)) {
 			// must be called while holding the lock - otherwise
 			// we have a race condition with solvers that
 			// are currently blocking on the mutex and we could enumerate 
 			// more models than requested by the user
-			terminateComplete(s.decisionLevel() == 0);
-		}
-		else if (!shared_->enumerator()->trivial(thread_[s.id()]->disjointPath())) {
-			// enumerator is not trivial w.r.t current search scheme
-			// force update of other solvers
-			postMessage(Message_t::update_message);
-			// ignore message in the active solver
-			thread_[s.id()]->setUpdate(msg_.updates);
+			terminate(s, r == Enumerator::enumerate_stop_complete);
 		}
 		if (shared_->enumerator()->enumerated == 1 && !shared_->enumerator()->supportsRestarts()) {
-			msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
-			msg_.setFlag(ThreadControlFlags::allow_gp_flag);
+			// switch to backtracking based splitting algorithm
+			shared_->setControl(SharedData::forbid_restart_flag | SharedData::allow_gp_flag);
 		}
 	}
-	return r == Enumerator::enumerate_continue && !msg_.terminate();
-}
-
-// checks for new update messages
-// POST: m.doUpdate if there is an active update message for m
-// return: m.doUpdate
-bool ParallelSolve::checkUpdates(Message_t::Local& m) const {
-	uint32 updates     = msg_.updates;
-	uint32 seen        = m.update;
-	m.update           = updates;
-	return (m.doUpdate = (m.doUpdate || seen != updates));
+	if (r == Enumerator::enumerate_continue) { 
+		// continue only after we released the mutex to avoid deadlock -
+		// continueFromModel(s) might call s.propagate() which could
+		// block on a sync message.
+		r = shared_->enumerator()->continueFromModel(s); 
+	}
+	return r == Enumerator::enumerate_continue && !shared_->terminate();
 }
 
 // updates s with new messages and uses s to create a new guiding path
 // if necessary and possible
-bool ParallelSolve::handleMessages(Solver& s, Message_t::Local& m) {
+bool ParallelSolve::handleMessages(Solver& s) {
 	// check if there are new messages for s
+	if (!shared_->hasMessage()) {
+		// nothing to do
+		return true; 
+	}
 	uint32 hId         = s.id();
-	uint32 gt          = msg_.messages;
 	ParallelHandler* h = thread_[hId];
-	if (m.any != gt) {
-		if (msg_.hasControlMsg()) {
-			if (msg_.terminate()) {
-				LOG(s.id(), "TERMINATE message received!");
-				h->handleTerminateMessage();
-				s.stats.parallel->terminated = true;
-				s.setStopConflict();
-				return false;
-			}
-			if (msg_.restart() && waitOnRestart(s)) {
-				s.setStopConflict();
-				return false;
-			}
-			return true;
-		}
-		checkUpdates(m);
-		uint32 mt  = msg_.workReq; 
-		if (mt <= 0) {
-			m.any    = gt;
-		}
-		else if (h->splittable()) {
-			// First declare split request as handled
-			// and only then do the actual split.
-			// This way, we minimize the chance for 
-			// "over"-splitting, i.e. one split request handled
-			// by more than one thread.
-			if (--msg_.workReq == 0) {
-				m.any  = gt;
-			}
-			h->handleSplitMessage();
+	if (shared_->terminate()) {
+		LOG(s.id(), "TERMINATE message received!");
+		h->handleTerminateMessage();
+		s.stats.parallel->terminated = true;
+		s.setStopConflict();
+		return false;
+	}
+	if (shared_->synchronize()) {
+		if (waitOnSync(s)) {
+			s.setStopConflict();
+			return false;
 		}
+		return true;
+	}
+	if (s.splittable() && shared_->workReq > 0) {
+		// First declare split request as handled
+		// and only then do the actual split.
+		// This way, we minimize the chance for 
+		// "over"-splitting, i.e. one split request handled
+		// by more than one thread.
+		shared_->aboutToSplit();
+		h->handleSplitMessage();
 	}
 	return true;
 }
 
 void ParallelSolve::requestRestart() {
-	if (msg_.allowRestart() && ++msg_.restartReq == numThreads()) {
-		postMessage(Message_t::restart_message);
+	if (shared_->allowRestart() && ++shared_->restartReq == numThreads()) {
+		shared_->postMessage(SharedData::msg_sync_restart, true);
 	}
 }
 
@@ -743,7 +730,6 @@ void ParallelHandler::attach(Solver& s, const SolveParams& p) {
 	s.stats.enableParallelStats();
 	aggStats.enableParallelStats();
 	if (s.stats.jumps) { aggStats.enableJumpStats(); }
-	messageHandler_.msg = MessageHandler::MessageQ();
 	gp_.reset();
 	intTail_= 0;
 	error_  = 0;
@@ -778,11 +764,6 @@ void ParallelHandler::detach() {
 	}
 }
 
-// true if solver can be used to split-off a new guiding path
-bool ParallelHandler::splittable() const {
-	return solver_->splittable() && !messageHandler_.pendingUpdate();
-}
-
 void ParallelHandler::prepareForGP(const LitVec& out, uint64 restart, bool fromSplit) {
 	gp_.reset(restart, fromSplit);
 	aggStats.parallel->newGP(out.size());
@@ -837,36 +818,41 @@ bool ParallelHandler::simplify(Solver& s, bool sh) {
 	return false;
 }
 
-// integrate new information from models
-bool ParallelHandler::MessageHandler::doUpdate(Solver& s, const GP& gp) {
-	// Skip update of model while assumption literal is not yet assigned.
-	// This is necessary during hierarchical optimization because otherwise
-	// integrating a too strong optimum might irrevocably force the assumption literal
-	// which would defeat its purpose.
-	if (s.isTrue(s.sharedContext()->tagLiteral())) {
-		if (!s.sharedContext()->enumerator()->update(s, gp.split)) {
-			return false;
+bool ParallelHandler::integrateClauses(Solver& s) {
+	assert(!s.hasConflict() && &s == solver_);
+	SharedLiterals* buffer[30];
+	uint32          rec = s.sharedContext()->receive(s, buffer, 30);
+	if (rec != 0) {
+		uint32 intFlags   = ctrl()->integrateFlags();
+		ClauseInfo e; e.setType(Constraint_t::learnt_other);
+		for (uint32 i = 0; i != rec;) {
+			ClauseCreator::Result ret = ClauseCreator::integrate(s, buffer[i++], intFlags, e);
+			if (ret.local) { add(ret.local); }
+			if (!ret.ok)   { while (i != rec) { buffer[i++]->release(); } return false; }				
 		}
-		msg.doUpdate = false;
 	}
-	return true;	
+	return true;
 }
 
 bool ParallelHandler::propagateFixpoint(Solver& s) {
-	if (messageHandler_.pendingUpdate() && !messageHandler_.doUpdate(s, gp_)) {
-		return false;
+	// Periodically check for updates from new models.
+	// Skip update while assumption literal is not yet assigned.
+	// This is necessary during hierarchical optimization because otherwise
+	// integrating a too strong optimum might irrevocably force the assumption literal
+	// which would defeat its purpose.
+	if ( (s.stats.choices & 63) == 0 && s.isTrue(s.sharedContext()->tagLiteral()) ) {
+		if (!s.sharedContext()->enumerator()->update(s, gp_.split)) {
+			return false;
+		}
+		if (s.queueSize() != 0 && !s.propagateUntil(this)) {
+			return false;
+		}
+	}
+	for (;;) {
+		if (!integrateClauses(s))    return false;
+		if (s.queueSize() == 0)      break;
+		if (!s.propagateUntil(this)) return false;
 	}
-	SharedLiterals* temp[5];
-	uint32 rec;
-	do {
-		rec = s.sharedContext()->receive(s, temp, 5);
-		if (!integrateClauses(s, temp, rec)) {
-			return false;
-		}
-		if (s.queueSize() > 0 && !s.propagateUntil(this)) {
-			return false;
-		}
-	} while (rec == 5);
 	if (s.stats.conflicts >= gp_.restart) {
 		ctrl()->requestRestart();
 		gp_.restart *= 2;
@@ -879,23 +865,10 @@ bool ParallelHandler::propagateFixpoint(Solver& s) {
 bool ParallelHandler::isModel(Solver& s) {
 	assert(s.numFreeVars() == 0);
 	// either no unprocessed updates or still a model after
-	// updates was integrated
-	return !ctrl()->checkUpdates(messageHandler_.msg)
-		|| (messageHandler_.doUpdate(s, gp_) && s.numFreeVars() == 0);
-}
-
-bool ParallelHandler::integrateClauses(Solver& s, SharedLiterals** arr, uint32 num) {
-	assert(!s.hasConflict() && &s == solver_);
-	ClauseCreator::Result ret(0, true);
-	ClauseInfo e; e.setType(Constraint_t::learnt_other);
-	uint32 intFlags = ctrl()->integrateFlags();
-	for (uint32 i = 0; i != num && ret.ok; ++i) {
-		ret = ClauseCreator::integrate(s, arr[i], intFlags, e);
-		if (ret.local) {
-			add(ret.local);
-		}
-	}
-	return ret.ok;
+	// updates were integrated
+	return s.sharedContext()->enumerator()->update(s, gp_.split)
+		&& s.numFreeVars() == 0
+		&& s.queueSize()   == 0;
 }
 
 void ParallelHandler::add(ClauseHead* h) {
