@@ -42,100 +42,91 @@ Literal SelectFirst::doSelect(Solver& s) {
 	return Literal();
 }
 /////////////////////////////////////////////////////////////////////////////////////////
-// SelectRandom selection strategy
-/////////////////////////////////////////////////////////////////////////////////////////
-// Selects a random literal from all free literals.
-class SelectRandom : public DecisionHeuristic {
-public:
-	SelectRandom() : randFreq_(1.0), pos_(0) {}
-	void shuffle(RNG& rng) {
-		std::random_shuffle(vars_.begin(), vars_.end(), rng);
-		pos_ = 0;
-	}
-	void   randFreq(double d) { randFreq_ = d; }
-	double randFreq() const   { return randFreq_; }
-	void   endInit(Solver& s) {
-		vars_.clear();
-		for (Var i = 1; i <= s.numVars(); ++i) {
-			if (s.value( i ) == value_free) {
-				vars_.push_back(i);
-			}
-		}
-		pos_ = 0;
-	}
-private:
-	Literal doSelect(Solver& s) {
-		LitVec::size_type old = pos_;
-		do {
-			if (s.value(vars_[pos_]) == value_free) {
-				Literal l = savedLiteral(s, vars_[pos_]);
-				return l != posLit(0)
-					? l
-					: s.sharedContext()->preferredLiteralByType(vars_[pos_]);
-			}
-			if (++pos_ == vars_.size()) pos_ = 0;
-		} while (old != pos_);
-		assert(!"SelectRandom::doSelect() - precondition violated!\n");
-		return Literal();
-	}
-	VarVec            vars_;
-	double            randFreq_;
-	VarVec::size_type pos_;
-};
-/////////////////////////////////////////////////////////////////////////////////////////
 // Post propagator list
 /////////////////////////////////////////////////////////////////////////////////////////
-Solver::PPList::PPList() : head(0), look(0), saved(0) { }
+Solver::PPList::PPList() : head(0), look(0) { }
 Solver::PPList::~PPList() {
+	enableFunky();
 	for (PostPropagator* r = head; r;) {
 		PostPropagator* t = r;
 		r = r->next;
 		t->destroy();
 	}
 }
+PostPropagator* Solver::PPList::disableFunky() {
+	// already disabled
+	if (look >= nil()) { return look != nil() ? look : 0; }
+	// move post propagators with priority >= lookahead to ext
+	PostPropagator* ext = 0;
+	if (head && head->priority() < PostPropagator::priority_lookahead) {
+		for (PostPropagator* r = head; r->next; r = r->next) {
+			if (r->next->priority() >= PostPropagator::priority_lookahead) {
+				ext     = r->next;
+				r->next = 0;
+				break;
+			}
+		}
+	}
+	else {
+		ext  = head;
+		head = 0;
+	}
+	// mark as disabled even if there is currently no funky post propagator
+	look = ext ? ext : nil();
+	return ext;
+}
+PostPropagator* Solver::PPList::enableFunky() {
+	// already enabled or empty list
+	if (look <= nil()) { return (look = 0); }
+	// move post propagators with priority >= lookahead back to active list
+	PostPropagator* ext = look;
+	look                = 0; // mark as enabled
+	PostPropagator* r  = head;
+	if (!r)         { return (head = ext); }
+	while (r->next) { r= r->next; }
+	return (r->next = ext);	
+}
+
 void Solver::PPList::add(PostPropagator* p) {
 	assert(p && p->next == 0);
 	uint32 prio = p->priority();
-	if (!head || prio == PostPropagator::priority_highest || prio < head->priority()) {
-		p->next = head;
-		head    = p;
+	PostPropagator** listHead = (look == 0 || prio < PostPropagator::priority_lookahead)
+		? &head
+		: &look;
+	if (*listHead <= nil() || prio < (*listHead)->priority()) {
+		p->next  = *listHead > nil() ? *listHead : 0;
+		*listHead= p;
+		return;
 	}
-	else {
-		for (PostPropagator* r = head; ; r = r->next) {
-			if (r->next == 0) {
-				r->next = p; break;
-			}
-			else if (prio < r->next->priority()) {
-				p->next = r->next;
-				r->next = p;
-				break;
-			}				
+	for (PostPropagator* r = *listHead;; r = r->next) {
+		if (!r->next || prio < r->next->priority()) {
+			p->next = r->next;
+			r->next = p;
+			return;
 		}
 	}
-	if (prio >= PostPropagator::priority_lookahead && (!look || prio < look->priority())) {
-		look = p;
-	}
-	saved = head;
 }
+
 void Solver::PPList::remove(PostPropagator* p) {
 	assert(p);
-	if (!head) return;
-	if (p == head) {
-		head = head->next;
-		p->next = 0;
+	PostPropagator** listHead = (look == 0 || p->priority() < PostPropagator::priority_lookahead)
+		? &head
+		: &look;
+	if (*listHead <= nil()) { return; }
+	if (*listHead == p)     {
+		*listHead = (*listHead)->next;
+		p->next   = 0;
+		return;
 	}
-	else {
-		for (PostPropagator* r = head; ; r = r->next) {
-			if (r->next == p) {
-				r->next = r->next->next;
-				p->next = 0;
-				break;
-			}
+	for (PostPropagator* r = *listHead; r->next; r = r->next) {
+		if (r->next == p) {
+			r->next = r->next->next;
+			p->next = 0;
+			return;
 		}
 	}
-	if (p == look) { look = look->next; }
-	saved = head;
 }
+
 bool Solver::PPList::propagate(Solver& s, PostPropagator* x) {
 	for (PostPropagator* p = head, *t; p != x;) {
 		// just in case t removes itself from the list
@@ -155,8 +146,8 @@ void Solver::PPList::simplify(Solver& s, bool shuf) {
 		}
 	}
 }
-bool Solver::PPList::init(Solver& s) {
-	for (PostPropagator* r = head; r; r = r->next) {
+bool Solver::PPList::init(Solver& s, PostPropagator* list) {
+	for (PostPropagator* r = list; r; r = r->next) {
 		if (!r->init(s)) return false;
 	}
 	return true;
@@ -170,25 +161,19 @@ bool PostPropagator::propagateFixpoint(Solver& s) {
 }
 void Solver::PPList::reset()            { for (PostPropagator* r = head; r; r = r->next) { r->reset(); } }
 bool Solver::PPList::isModel(Solver& s) {
-	saved = head;
+	if (s.hasConflict()) { return false; }
 	for (PostPropagator* r = head; r; r = r->next) {
 		if (!r->isModel(s)) return false;
 	}
 	return true;
 }
-bool Solver::PPList::nextSymModel(Solver& s, bool expand) {
-	for (; saved; saved = saved->next) {
-		if (saved->nextSymModel(s, expand)) return true;
-	}
-	saved = head;
-	return false;
-}
 /////////////////////////////////////////////////////////////////////////////////////////
 // SolverStrategies
 /////////////////////////////////////////////////////////////////////////////////////////
 SolverStrategies::SolverStrategies()
 	: heuristic(new SelectFirst) 
 	, search(use_learning)
+	, heuOpts(UINT32_MAX)
 	, saveProgress(0) 
 	, reverseArcs(0)
 	, otfs(0)
@@ -209,7 +194,6 @@ Solver::Solver()
 	, ccMin_(0)
 	, smallAlloc_(new SmallClauseAlloc)
 	, shared_(0)
-	, randHeuristic_(0)
 	, levConflicts_(0)
 	, undoHead_(0)
 	, enum_(0)
@@ -251,11 +235,9 @@ void Solver::freeMem() {
 	delete smallAlloc_;
 	delete ccMin_;
 	delete levConflicts_;
-	delete randHeuristic_;
 	smallAlloc_   = 0;
 	ccMin_        = 0;
 	levConflicts_ = 0;
-	randHeuristic_= 0;
 }
 
 SatPreprocessor* Solver::satPrepro() const { return shared_->satPrepro.get(); }
@@ -283,19 +265,24 @@ void Solver::startInit(SharedContext* ctx, uint32 numConsGuess) {
 			undoFree(new ConstraintDB(10)); 
 		}
 	}
+	if (strategy_.heuOpts == UINT32_MAX) {
+		strategy_.heuOpts = (id() & 1);
+	}
 	strategy_.heuristic->startInit(*this);
+	// disable any funky post propagators during setup
+	post_.disableFunky();
 }
 
 bool Solver::endInit() {
 	if (!propagate()) { return false; }
 	strategy_.heuristic->endInit(*this);
-	// Force propagation again - just in case
-	// heuristic->endInit() added new information
-	if (!propagate() || !simplify()) {
+	// enable funky post propagators and
+	// force initial propagation
+	PostPropagator* ext = post_.enableFunky();
+	if (ext && (!post_.init(*this, ext) || !propagate())) {
 		return false;
 	}
-	if (randHeuristic_) randHeuristic_->endInit(*this);
-	return true;
+	return simplify();
 }
 
 bool Solver::addUnary(Literal p, ConstraintType t) {
@@ -482,21 +469,6 @@ void Solver::removeUndoWatch(uint32 dl, Constraint* c) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // Solver: Basic DPLL-functions
 ////////////////////////////////////////////////////////////////////////////////////////
-void Solver::initRandomHeuristic(double randFreq) {
-	randFreq = std::min(1.0, std::max(0.0, randFreq));
-	if (randFreq == 0.0) {
-		delete randHeuristic_;
-		randHeuristic_ = 0;
-		return;
-	}
-	if (!randHeuristic_) {
-		randHeuristic_ = new SelectRandom();
-		randHeuristic_->endInit(*this);
-	}
-	static_cast<SelectRandom*>(randHeuristic_)->shuffle(strategies().rng);
-	static_cast<SelectRandom*>(randHeuristic_)->randFreq(randFreq);
-}
-
 
 // removes all satisfied binary and ternary clauses as well
 // as all constraints for which Constraint::simplify returned true.
@@ -607,6 +579,7 @@ void Solver::simplifyShort(Literal p) {
 }
 
 void Solver::setConflict(Literal p, const Antecedent& a, uint32 data) {
+	++stats.conflicts;
 	conflict_.push_back(~p);
 	if (strategy_.search != SolverStrategies::no_learning && !a.isNull()) {
 		if (data == UINT32_MAX) {
@@ -629,7 +602,7 @@ bool Solver::assume(const Literal& p) {
 		assert(decisionLevel() != assign_.maxLevel());
 		++stats.choices;
 		levels_.push_back(DLevel(numAssignedVars(), 0));
-		if (levConflicts_) levConflicts_->push_back( (uint32)stats.conflicts );
+		if (levConflicts_) levConflicts_->push_back( 0 );
 		return assign_.assign(p, decisionLevel(), Antecedent());
 	}
 	return isTrue(p);
@@ -722,11 +695,11 @@ bool Solver::unitPropagate() {
 	return decisionLevel() > 0 || (units_=mark(units_, uint32(assign_.front))) == assign_.front;
 }
 
-bool Solver::test(Literal p, Constraint* c) {
+bool Solver::test(Literal p, PostPropagator* c) {
 	assert(value(p.var()) == value_free && !hasConflict());
 	assume(p); --stats.choices;
 	freezeLevel(decisionLevel()); // can't split-off this level
-	if (unitPropagate() && post_.propagate(*this, post_.look)) {
+	if (propagateUntil(c)) {
 		assert(decision(decisionLevel()) == p);
 		unfreezeLevel(decisionLevel());
 		if (c) c->undoLevel(*this);
@@ -742,7 +715,6 @@ bool Solver::test(Literal p, Constraint* c) {
 
 bool Solver::resolveConflict() {
 	assert(hasConflict());
-	++stats.conflicts;
 	if (decisionLevel() > rootLevel_) {
 		if (decisionLevel() != btLevel_ && strategy_.search != SolverStrategies::no_learning) {
 			uint32 uipLevel = analyzeConflict();
@@ -754,9 +726,6 @@ bool Solver::resolveConflict() {
 			return backtrack();
 		}
 	}
-	// do not count artificial conflicts that are only used
-	// to stop the current search
-	stats.conflicts -= hasStopConflict();
 	return false;
 }
 
@@ -864,7 +833,10 @@ void Solver::undoLevel(bool sp) {
 		undoFree(levels_.back().undo);
 	}
 	levels_.pop_back();
-	if (levConflicts_) levConflicts_->pop_back();
+	if (levConflicts_) {
+		(*levConflicts_)[levels_.size()] += levConflicts_->back();
+		levConflicts_->pop_back();
+	}
 }
 
 // computes the First-UIP clause and stores it in cc_, where cc_[0] is the asserting literal (inverted UIP)
@@ -912,6 +884,7 @@ uint32 Solver::analyzeConflict() {
 			// resolvent subsumes rhs and possibly also lhs
 			otfs(lhs, rhs, p, onLevel == 1);
 		}
+		assert(onLevel > 0 && "CONFLICT MUST BE ANALYZED ON CONFLICT LEVEL!");
 		// search for the last assigned literal that needs to be analyzed...
 		while (!seen(assign_.last().var())) {
 			assign_.undoLast();
@@ -1206,18 +1179,25 @@ bool Constraint::minimize(Solver& s, Literal p, CCMinRecursive* rec) {
 	return true;
 }
 
-
-// Selects next branching literal. Use user-supplied heuristic if rand() < randProp.
-// Otherwise makes a random choice.
-// Returns false if assignment is total.
-bool Solver::decideNextBranch() {
-	DecisionHeuristic* heu = strategy_.heuristic.get();
-	if (randHeuristic_ && strategy_.rng.drand() < static_cast<SelectRandom*>(randHeuristic_)->randFreq()) {
-		heu = randHeuristic_;
+// Selects next branching literal
+bool Solver::decideNextBranch(double f) { 
+	if (f <= 0.0 || strategy_.rng.drand() >= f || numFreeVars() == 0) {
+		return strategy_.heuristic.get()->select(*this);
 	}
-	return heu->select(*this);
+	// select randomly
+	Literal choice;
+	uint32 maxVar = numVars() + 1; 
+	for (uint32 v = strategy_.rng.irand(maxVar);;) {
+		if (value(v) == value_free) {
+			choice    = savedValue(v) == value_free
+				? sharedContext()->preferredLiteralByType(v)
+				: Literal(v, savedValue(v) == value_false);
+			break;
+		}
+		if (++v == maxVar) { v = 1; }
+	}
+	return assume(choice);
 }
-
 // Removes up to maxRem% of the learnt nogoods but
 // keeps those that are locked or are highly active.
 void Solver::reduceLearnts(float maxRem) {
@@ -1380,66 +1360,45 @@ uint32 Solver::computeLbd(Literal p, const Literal* first, const Literal* last)
 /////////////////////////////////////////////////////////////////////////////////////////
 // The basic DPLL-like search-function
 /////////////////////////////////////////////////////////////////////////////////////////
-ValueRep Solver::search(RestartLimit& rlimit, LearntLimit& dlimit, double randProp) {
-	initRandomHeuristic(randProp);
-	uint64 cfl     = 0; // number of conflicts
-	rlimit.maxConf = std::max(uint64(1), rlimit.maxConf);
-	if (rlimit.countLocal) {
+ValueRep Solver::search(SearchLimits& limit, double rf) {
+	rf            = std::max(0.0, std::min(1.0, rf));
+	VarVec* local = 0;
+	if (limit.local != UINT64_MAX) {
 		if (!levConflicts_) levConflicts_ = new VarVec();
 		if (decisionLevel() == rootLevel()) { levConflicts_->clear(); }
-		levConflicts_->resize(decisionLevel()+1, (uint32)stats.conflicts);
+		levConflicts_->resize(decisionLevel()+1, 0);
+		local = levConflicts_;
 	}
-	do {
-		if ((hasConflict()&&!resolveConflict()) || !simplify()) { return value_false; }
-		do {
-			while (!propagate()) {
-				++cfl;
-				if (!resolveConflict() || (decisionLevel() == 0 && !simplify())) {
-					return value_false;
-				}
-				if ( (cfl == rlimit.maxConf && !rlimit.countLocal) ||
-					(rlimit.countLocal && (stats.conflicts - (*levConflicts_)[decisionLevel()]) >= rlimit.maxConf)) {
-					rlimit.maxConf = 0;
-					dlimit.maxConf-= std::min(cfl, dlimit.maxConf);
-					return value_free;  
-				}
+	unconstr_.clear();
+BEGIN_SEARCH:
+	for (uint32 conflicts = hasConflict() || !propagate() || !simplify();;) {
+		if (conflicts) {
+			for (; resolveConflict() && !propagate(); ) {
+				++conflicts;
 			}
-			if (numLearntConstraints()>dlimit.maxLearnt || cfl >= dlimit.maxConf) { 
-				dlimit.maxConf -= std::min(cfl, dlimit.maxConf);
-				rlimit.maxConf -= !rlimit.countLocal ? cfl : 0;
-				return value_free; // dlimit reached
-			}
-		} while (decideNextBranch());
-		// found a model candidate
-		assert(numFreeVars() == 0 || hasConflict());
-	} while (hasConflict() || !post_.isModel(*this));
+			limit.conflicts -= std::min<uint64>(conflicts, limit.conflicts);
+			if (local && (local->back()+=conflicts) >= limit.local)        { limit.local = 0;    }
+			if (hasConflict() || (decisionLevel() == 0 && !simplify()))    { return value_false; }
+			if (numLearntConstraints() > limit.learnts || limit.reached()) { return value_free;  }
+		}
+		if      (decideNextBranch(rf)) { conflicts = !propagate(); }
+		else if (post_.isModel(*this)) { break;             }
+		else                           { goto BEGIN_SEARCH; }
+	}
 	stats.addModel(decisionLevel());
 	if (satPrepro()) {
 		satPrepro()->extendModel(assign_, unconstr_);
 	}
-	dlimit.maxConf -= cfl;
-	rlimit.maxConf -= !rlimit.countLocal ? cfl : 0;
 	return value_true;
 }
 
-bool Solver::nextSymModel(bool expand) {
+bool Solver::nextSymModel() {
 	assert(numFreeVars() == 0);
-	if (expand) {
-		if (satPrepro() != 0 && !unconstr_.empty()) {
-			stats.addModel(decisionLevel());
-			satPrepro()->extendModel(assign_, unconstr_);
-			return true;
-		}
-		else if (post_.nextSymModel(*this, true)) {
-			stats.addModel(decisionLevel());
-			return true;
-		}
-	}
-	else {
-		post_.nextSymModel(*this, false);
-		unconstr_.clear();
+	if (satPrepro() && !unconstr_.empty()) {
+		stats.addModel(decisionLevel());
+		satPrepro()->extendModel(assign_, unconstr_);
+		return true;
 	}
 	return false;
 }
-
 }
