@@ -56,7 +56,7 @@ public:
 		MinimizeBuilder b;
 		builder.addMinimize(b);
 		ctx.addEnumerator(e.release());
-		ctx.enumerator()->setMinimize(b.build(ctx), false);
+		ctx.enumerator()->setMinimize(b.build(ctx));
 		ctx.endInit(1);
 		SymbolTable& index = ctx.symTab();
 		solver.assume(index[1].lit);
