@@ -680,16 +680,15 @@ public:
 	}
 private:
 	const char* next() {
-		while (end_ != cmd_.size() && std::isspace(static_cast<unsigned char>(cmd_[end_]))) {
-			++end_;
-		}
-		if ((pos_ = end_) >= cmd_.size()) return 0;
-		while (end_ != cmd_.size() && !std::isspace(static_cast<unsigned char>(cmd_[end_]))) {
-			++end_;
-		}
-		if (end_ != cmd_.size()) {
-			cmd_[end_++] = '\0';
-		}
+		const std::string::size_type eos = cmd_.size();
+		// skip any leading whitespace
+		while (end_ != eos && std::isspace(static_cast<unsigned char>(cmd_[end_])))  { ++end_; }
+		if (end_ == eos) { return 0; }
+		// find end of current arg
+		pos_ = end_++;
+		while (end_ != eos && !std::isspace(static_cast<unsigned char>(cmd_[end_]))) { ++end_; }
+		// terminate and return current arg
+		if (end_ != eos) { cmd_[end_++] = '\0'; }
 		return cmd_.c_str() + pos_;
 	}
 	CommandStringParser& operator=(const CommandStringParser&);
